"""
Module for isochrone data from evolutionary models.
"""

from typing import Optional

import h5py

from species.data.isochrone_data.iso_ames import add_ames
from species.data.isochrone_data.iso_atmo import add_atmo
from species.data.isochrone_data.iso_baraffe2015 import add_baraffe2015
from species.data.isochrone_data.iso_btsettl import add_btsettl
from species.data.isochrone_data.iso_linder2019 import add_linder2019
from species.data.isochrone_data.iso_manual import add_manual
from species.data.isochrone_data.iso_marleau import add_marleau
from species.data.isochrone_data.iso_nextgen import add_nextgen
from species.data.isochrone_data.iso_saumon2008 import add_saumon2008
from species.data.isochrone_data.iso_sonora import add_sonora


def add_isochrone_grid(
    data_folder: str,
    hdf5_file: h5py._hl.files.File,
    model_name: str,
    filename: Optional[str] = None,
    tag: Optional[str] = None,
) -> None:
    """
    Function for adding an isochrone grid to the database.

    Parameters
    ----------
    input_path : str
        Folder where the data is located.
    database : h5py._hl.files.File
        Database.
    model_name : str
        Evolutionary model ('ames', 'atmo', 'baraffe2015',
        'bt-settl', 'linder2019', 'nextgen', 'saumon2008',
        'sonora', or 'manual'). Isochrones will be
        automatically downloaded. Alternatively,
        isochrone data can be downloaded from
        https://phoenix.ens-lyon.fr/Grids/ or
        https://perso.ens-lyon.fr/isabelle.baraffe/, and can
        be manually added by setting the ``filename`` and
        ``tag`` arguments, and setting ``model='manual'``.
    filename : str, None
        Filename with the isochrone data. Setting the argument
        is only required when ``model='manual'``. Otherwise,
        the argument can be set to ``None``.
    tag : str, None
        Database tag name where the isochrone that will be
        stored. Setting the argument is only required when
        ``model='manual'``. Otherwise, the argument can be
        set to ``None``.

    Returns
    -------
    None
        NoneType
    """

    if model_name == "ames":
        add_ames(hdf5_file, data_folder)

    elif model_name == "atmo":
        add_atmo(hdf5_file, data_folder)

    elif model_name == "baraffe2015":
        add_baraffe2015(hdf5_file, data_folder)

    elif model_name == "bt-settl":
        add_btsettl(hdf5_file, data_folder)

    elif model_name == "linder2019":
        add_linder2019(hdf5_file, data_folder)

    elif model_name == "manual":
        add_manual(hdf5_file, tag, filename)

    elif model_name == "marleau":
        add_marleau(hdf5_file, tag, filename)

    elif model_name == "nextgen":
        add_nextgen(hdf5_file, data_folder)

    elif model_name == "saumon2008":
        add_saumon2008(hdf5_file, data_folder)

    elif model_name == "sonora":
        add_sonora(hdf5_file, data_folder)

    else:
        raise ValueError(
            f"The evolutionary model_name '{model_name}' is "
            "not supported. Please choose another "
            "argument for 'model_name'. Have a look "
            "at the documentation of add_isochrones "
            "for details on the supported model_names."
        )



"""
Utility functions for model spectra.
"""

import warnings

from typing import Dict, Tuple, Union

import numpy as np

from typeguard import typechecked

from species.core.box import ModelBox, create_box
from species.util.spec_util import create_wavelengths


@typechecked
def convert_model_name(in_name: str) -> str:
    """
    Function for updating a model name for use in plots.

    Parameters
    ----------
    in_name : str
        Model name as used by species.

    Returns
    -------
    str
        Updated model name for plots.
    """

    if in_name == "drift-phoenix":
        out_name = "DRIFT-PHOENIX"

    elif in_name == "ames-cond":
        out_name = "AMES-Cond"

    elif in_name == "ames-dusty":
        out_name = "AMES-Dusty"

    elif in_name == "atmo":
        out_name = "ATMO"

    elif in_name == "atmo-ceq":
        out_name = "ATMO CEQ"

    elif in_name == "atmo-neq-weak":
        out_name = "ATMO NEQ weak"

    elif in_name == "atmo-neq-strong":
        out_name = "ATMO NEQ strong"

    elif in_name == "petrus2023":
        out_name = "ATMO (Petrus et al. 2023)"

    elif in_name == "atmo-petrus2023":
        out_name = "ATMO (Petrus et al. 2023)"

    elif in_name == "bt-cond":
        out_name = "BT-Cond"

    elif in_name == "bt-cond-feh":
        out_name = "BT-Cond"

    elif in_name == "bt-settl":
        out_name = "BT-Settl"

    elif in_name == "bt-settl-cifist":
        out_name = "BT-Settl"

    elif in_name == "bt-nextgen":
        out_name = "BT-NextGen"

    elif in_name == "bt-nextgen-subsolar":
        out_name = "BT-NextGen"

    elif in_name == "koester-wd":
        out_name = "Koester (2010)"

    elif in_name == "petitcode-cool-clear":
        out_name = "petitCODE clear"

    elif in_name == "petitcode-cool-cloudy":
        out_name = "petitCODE cloudy"

    elif in_name == "petitcode-hot-clear":
        out_name = "petitCODE clear"

    elif in_name == "petitcode-hot-cloudy":
        out_name = "petitCODE cloudy"

    elif in_name == "petitcode-linder2019-clear":
        out_name = "petitCODE clear (Linder et al. 2019)"

    elif in_name == "petitcode-linder2019-cloudy":
        out_name = "petitCODE cloudy (Linder et al. 2019)"

    elif in_name == "exo-rem":
        out_name = "Exo-REM"

    elif in_name == "exo-rem-highres":
        out_name = "Exo-REM"

    elif in_name in ["planck", "blackbody"]:
        out_name = "Blackbody"

    elif in_name == "zhu2015":
        out_name = "Zhu (2015)"

    elif in_name == "saumon2008-clear":
        out_name = "Saumon & Marley (2008)"

    elif in_name == "saumon2008-cloudy":
        out_name = "Saumon & Marley (2008)"

    elif in_name == "sonora-cholla":
        out_name = "Sonora Cholla"

    elif in_name == "sonora-bobcat":
        out_name = "Sonora Bobcat"

    elif in_name == "sonora-bobcat-co":
        out_name = "Sonora Bobcat C/O"

    elif in_name == "sonora-elfowl-l":
        out_name = "Sonora Elf Owl"

    elif in_name == "sonora-elfowl-t":
        out_name = "Sonora Elf Owl"

    elif in_name == "sonora-elfowl-y":
        out_name = "Sonora Elf Owl"

    elif in_name == "sphinx":
        out_name = "SPHINX"

    elif in_name == "petitradtrans":
        out_name = "petitRADTRANS"

    else:
        out_name = in_name

        warnings.warn(
            f"The model name '{in_name}' is not known "
            "so the output name will not get adjusted "
            "for plot purposes"
        )

    return out_name


@typechecked
def powerlaw_spectrum(
    wavel_range: Union[Tuple[float, float], Tuple[np.float32, np.float32]],
    model_param: Dict[str, float],
    spec_res: float = 100.0,
) -> ModelBox:
    """
    Function for calculating a power-law spectrum. The power-law
    function is calculated in log(wavelength)-log(flux) space but
    stored in the :class:`~species.core.box.ModelBox` in linear
    wavelength-flux space.

    Parameters
    ----------
    wavel_range : tuple(float, float)
        Tuple with the minimum and maximum wavelength (um).
    model_param : dict
        Dictionary with the model parameters. Should contain
        `'log_powerlaw_a'`, `'log_powerlaw_b'`, and `'log_powerlaw_c'`.
    spec_res : float
        Spectral resolution (default: 100).

    Returns
    -------
    species.core.box.ModelBox
        Box with the power-law spectrum.
    """

    wavel = create_wavelengths((wavel_range[0], wavel_range[1]), spec_res)

    wavel *= 1e3  # (um) -> (nm)

    log_flux = (
        model_param["log_powerlaw_a"]
        + model_param["log_powerlaw_b"]
        * np.log10(wavel) ** model_param["log_powerlaw_c"]
    )

    model_box = create_box(
        boxtype="model",
        model="powerlaw",
        wavelength=1e-3 * wavel,  # (um)
        flux=10.0**log_flux,  # (W m-2 um-1)
        parameters=model_param,
        quantity="flux",
    )

    return model_box


@typechecked
def gaussian_spectrum(
    wavel_range: Union[Tuple[float, float], Tuple[np.float32, np.float32]],
    model_param: Dict[str, float],
    spec_res: float = 100.0,
    double_gaussian: bool = False,
) -> ModelBox:
    """
    Function for calculating a Gaussian spectrum (i.e. for an emission
    line).

    Parameters
    ----------
    wavel_range : tuple(float, float)
        Tuple with the minimum and maximum wavelength (um).
    model_param : dict
        Dictionary with the model parameters. Should contain
        ``'gauss_amplitude'``, ``'gauss_mean'``, ``'gauss_sigma'``,
        and optionally ``'gauss_offset'``.
    spec_res : float
        Spectral resolution (default: 100).
    double_gaussian : bool
        Set to ``True`` for returning a double Gaussian function.
        In that case, ``model_param`` should also contain
        ``'gauss_amplitude_2'``, ``'gauss_mean_2'``, and
        ``'gauss_sigma_2'``.

    Returns
    -------
    species.core.box.ModelBox
        Box with the Gaussian spectrum.
    """

    wavel = create_wavelengths((wavel_range[0], wavel_range[1]), spec_res)

    gauss_exp = np.exp(
        -0.5
        * (wavel - model_param["gauss_mean"]) ** 2
        / model_param["gauss_sigma"] ** 2
    )

    flux = model_param["gauss_amplitude"] * gauss_exp

    if double_gaussian:
        gauss_exp = np.exp(
            -0.5
            * (wavel - model_param["gauss_mean_2"]) ** 2
            / model_param["gauss_sigma_2"] ** 2
        )

        flux += model_param["gauss_amplitude_2"] * gauss_exp

    if "gauss_offset" in model_param:
        flux += model_param["gauss_offset"]

    model_box = create_box(
        boxtype="model",
        model="gaussian",
        wavelength=wavel,
        flux=flux,
        parameters=model_param,
        quantity="flux",
    )

    return model_box


# @typechecked
# def add_luminosity(modelbox: ModelBox) -> ModelBox:
#     """
#     Function to add the luminosity of a model spectrum to the parameter
#     dictionary of the box.
#
#     Parameters
#     ----------
#     modelbox : species.core.box.ModelBox
#         Box with the model spectrum. Should also contain the dictionary
#         with the model parameters, the radius in particular.
#
#     Returns
#     -------
#     species.core.box.ModelBox
#         The input box with the luminosity added in the parameter
#         dictionary.
#     """
#
#     print("Calculating the luminosity...", end="", flush=True)
#
#     if modelbox.model == "planck":
#         readmodel = ReadPlanck(wavel_range=(1e-1, 1e3))
#         fullspec = readmodel.get_spectrum(
#             model_param=modelbox.parameters, spec_res=1000.0
#         )
#
#     else:
#         readmodel = ReadModel(modelbox.model)
#         fullspec = readmodel.get_model(modelbox.parameters)
#
#     flux = simps(fullspec.flux, fullspec.wavelength)
#
#     if "parallax" in modelbox.parameters:
#         luminosity = (
#             4.0
#             * np.pi
#             * (1e3 * constants.PARSEC / fullspec.parameters["parallax"]) ** 2
#             * flux
#         )  # (W)
#
#     elif "distance" in modelbox.parameters:
#         luminosity = (
#             4.0
#             * np.pi
#             * (fullspec.parameters["distance"] * constants.PARSEC) ** 2
#             * flux
#         )  # (W)
#
#         # Analytical solution for a single-component Planck function
#         # luminosity = 4.*np.pi*(modelbox.parameters['radius']*constants.R_JUP)**2* \
#         #     constants.SIGMA_SB*modelbox.parameters['teff']**4.
#
#     else:
#         luminosity = (
#             4.0 * np.pi * (fullspec.parameters["radius"] * constants.R_JUP) ** 2 * flux
#         )  # (W)
#
#     modelbox.parameters["luminosity"] = luminosity / constants.L_SUN  # (Lsun)
#
#     print(" [DONE]")
#
#     print(
#         f"Wavelength range (um): {fullspec.wavelength[0]:.2e} - "
#         f"{fullspec.wavelength[-1]:.2e}"
#     )
#
#     print(f"Luminosity (Lsun): {luminosity/constants.L_SUN:.2e}")
#
#     return modelbox


@typechecked
def binary_to_single(param_dict: Dict[str, float], star_index: int) -> Dict[str, float]:
    """
    Function for converting a dictionary with atmospheric parameters
    of a binary system to a dictionary of parameters for one of the
    two stars.

    Parameters
    ----------
    param_dict : dict
        Dictionary with the atmospheric parameters of both stars. The
        keywords end either with ``_0`` or ``_1`` that correspond with
        ``star_index=0`` or ``star_index=1``.
    star_index : int
        Star index (0 or 1) that is used for the parameters in
        ``param_dict``.

    Returns
    -------
    dict
        Dictionary with the parameters of the selected star.
    """

    new_dict = {}

    for key, value in param_dict.items():
        if star_index == 0 and key[-1] == "0":
            new_dict[key[:-2]] = value

        elif star_index == 1 and key[-1] == "1":
            new_dict[key[:-2]] = value

        elif key in [
            "teff",
            "logg",
            "feh",
            "c_o_ratio",
            "fsed",
            "radius",
            "distance",
            "parallax",
        ]:
            new_dict[key] = value

    return new_dict



"""
Module with reading functionalities for calibration spectra.
"""

import configparser
import os

from typing import Dict, Optional, Tuple

import h5py
import numpy as np
import spectres

from scipy import interpolate, optimize

from typeguard import typechecked

from species.core.box import SpectrumBox, create_box
from species.phot.syn_phot import SyntheticPhotometry
from species.read.read_filter import ReadFilter
from species.util.spec_util import create_wavelengths, smooth_spectrum


class ReadCalibration:
    """
    Class for reading a calibration spectrum from the database.
    """

    @typechecked
    def __init__(self, tag: str, filter_name: Optional[str] = None) -> None:
        """
        Parameters
        ----------
        tag : str
            Database tag of the calibration spectrum.
        filter_name : str, None
            Filter that is used for the wavelength range. The full
            spectrum is read if the argument is set to ``None``.

        Returns
        -------
        NoneType
            None
        """

        self.tag = tag
        self.filter_name = filter_name

        if filter_name is None:
            self.wavel_range = None

        else:
            transmission = ReadFilter(filter_name)
            self.wavel_range = transmission.wavelength_range()

        config_file = os.path.join(os.getcwd(), "species_config.ini")

        config = configparser.ConfigParser()
        config.read(config_file)

        self.database = config["species"]["database"]

    @typechecked
    def resample_spectrum(
        self,
        wavel_points: np.ndarray,
        model_param: Optional[Dict[str, float]] = None,
        spec_res: Optional[float] = None,
        apply_mask: bool = False,
        interp_highres: bool = False,
    ) -> SpectrumBox:
        """
        Function for resampling the spectrum and optional
        uncertainties onto a new wavelength grid.

        Parameters
        ----------
        wavel_points : np.ndarray
            Wavelengths (um).
        model_param : dict, None
            Dictionary with the model parameters, which should only
            contain the ``'scaling'`` keyword. No scaling is applied if
            the argument of ``model_param`` is set to ``None``.
        spec_res : float, None
            Spectral resolution that is used for smoothing the spectrum
            before resampling the wavelengths. No smoothing is applied
            if the argument is set to ``None``. The smoothing can only
            be applied to spectra with a constant spectral resolution
            (which is the case for all model spectra that are
            supported by ``species``) or a constant wavelength
            spacing. The first smoothing approach is fastest.
        apply_mask : bool
            Exclude negative values and NaNs.
        interp_highres : bool
            Oversample the spectrum to $R = 10000$, such that the
            ``spec_res`` parameter can be applied on a spectrum with
            constant $\\lambda/\\Delta\\lambda$. The uncertainties
            are crudely propagated with an interpolation as well
            and should only be considered as estimate.

        Returns
        -------
        species.core.box.SpectrumBox
            Box with the resampled spectrum.
        """

        calib_box = self.get_spectrum(apply_mask=apply_mask)

        if interp_highres:
            flux_interp = interpolate.interp1d(
                calib_box.wavelength,
                calib_box.flux,
                bounds_error=False,
                fill_value="extrapolate",
            )

            sigma_interp = interpolate.interp1d(
                calib_box.wavelength,
                calib_box.error,
                bounds_error=False,
                fill_value="extrapolate",
            )

            wavel_range = (calib_box.wavelength[0], calib_box.wavelength[-1])
            wavel_highres = create_wavelengths(wavel_range, 10000.0)

            calib_box.wavelength = wavel_highres
            calib_box.flux = flux_interp(wavel_highres)
            calib_box.error = sigma_interp(wavel_highres)

            if spec_res is not None:
                calib_box.flux = smooth_spectrum(
                    wavelength=calib_box.wavelength,
                    flux=calib_box.flux,
                    spec_res=spec_res,
                )

            flux_interp = interpolate.interp1d(
                calib_box.wavelength,
                calib_box.flux,
                bounds_error=False,
                fill_value="extrapolate",
            )

            sigma_interp = interpolate.interp1d(
                calib_box.wavelength,
                calib_box.error,
                bounds_error=False,
                fill_value="extrapolate",
            )

            flux_new = flux_interp(wavel_points)
            error_new = sigma_interp(wavel_points)

        else:
            if spec_res is not None:
                calib_box.flux = smooth_spectrum(
                    wavelength=calib_box.wavelength,
                    flux=calib_box.flux,
                    spec_res=spec_res,
                )

            flux_new, error_new = spectres.spectres(
                wavel_points,
                calib_box.wavelength,
                calib_box.flux,
                spec_errs=calib_box.error,
                fill=0.0,
                verbose=False,
            )

        if model_param is not None:
            flux_new = model_param["scaling"] * flux_new
            error_new = model_param["scaling"] * error_new

        return create_box(
            boxtype="spectrum",
            spectrum="calibration",
            wavelength=wavel_points,
            flux=flux_new,
            error=error_new,
            name=self.tag,
        )

    @typechecked
    def get_spectrum(
        self,
        model_param: Optional[Dict[str, float]] = None,
        apply_mask: bool = False,
        wavel_sampling: Optional[float] = None,
        extrapolate: bool = False,
        min_wavelength: Optional[float] = None,
    ) -> SpectrumBox:
        """
        Function for selecting the calibration spectrum.

        Parameters
        ----------
        model_param : dict, None
            Model parameters. Should contain the 'scaling' value. Not
            used if set to ``None``.
        apply_mask : bool
            Exclude negative values and NaN values.
        wavel_sampling : float, None
            Wavelength sampling :math:`\\lambda/\\Delta\\lambda`. The
            original wavelength points are used if set to ``None``.
        extrapolate : bool
            Extrapolate to 6 um by fitting a power law function.
        min_wavelength : float, None
            Minimum wavelength used for fitting the power law function.
            All data is used if set to ``None``.

        Returns
        -------
        species.core.box.SpectrumBox
            Box with the spectrum.
        """

        with h5py.File(self.database, "r") as h5_file:
            cal_spec = np.array(h5_file[f"spectra/calibration/{self.tag}"])

            wavelength = cal_spec[0,]
            flux = cal_spec[1,]
            error = cal_spec[2,]

        if apply_mask:
            indices = np.where(flux > 0.0)[0]

            wavelength = wavelength[indices]
            flux = flux[indices]
            error = error[indices]

        if model_param is not None:
            flux = model_param["scaling"] * flux
            error = model_param["scaling"] * error

        if self.wavel_range is None:
            wl_index = np.ones(wavelength.size, dtype=bool)
        else:
            wl_index = (
                (flux > 0.0)
                & (wavelength > self.wavel_range[0])
                & (wavelength < self.wavel_range[1])
            )

        count = np.count_nonzero(wl_index)

        if count > 0:
            index = np.where(wl_index)[0]

            if index[0] > 0:
                wl_index[index[0] - 1] = True

            if index[-1] < len(wl_index) - 1:
                wl_index[index[-1] + 1] = True

            wavelength = wavelength[wl_index]
            flux = flux[wl_index]
            error = error[wl_index]

        if extrapolate:

            def _power_law(wavelength, offset, scaling, power_index):
                return offset + scaling * wavelength**power_index

            if min_wavelength:
                indices = np.where(wavelength > min_wavelength)[0]
            else:
                indices = np.arange(0, wavelength.size, 1)

            popt, pcov = optimize.curve_fit(
                f=_power_law,
                xdata=wavelength[indices],
                ydata=flux[indices],
                p0=(0.0, np.mean(flux[indices]), -1.0),
                sigma=error[indices],
            )

            sigma = np.sqrt(np.diag(pcov))

            print("Fit result for f(x) = a + b*x^c:")
            print(f"a = {popt[0]} +/- {sigma[0]}")
            print(f"b = {popt[1]} +/- {sigma[1]}")
            print(f"c = {popt[2]} +/- {sigma[2]}")

            while wavelength[-1] <= 6.0:
                wl_add = wavelength[-1] + wavelength[-1] / 1000.0

                wavelength = np.append(wavelength, wl_add)
                flux = np.append(flux, _power_law(wl_add, popt[0], popt[1], popt[2]))
                error = np.append(error, 0.0)

        if wavel_sampling is not None:
            wavelength_new = create_wavelengths(
                (wavelength[0], wavelength[-1]), wavel_sampling
            )

            flux_new, error_new = spectres.spectres(
                wavelength_new,
                wavelength,
                flux,
                spec_errs=error,
                fill=0.0,
                verbose=True,
            )

            wavelength = wavelength_new
            flux = flux_new
            error = error_new

        return create_box(
            boxtype="spectrum",
            spectrum="calibration",
            wavelength=wavelength,
            flux=flux,
            error=error,
            name=self.tag,
        )

    @typechecked
    def get_flux(
        self, model_param: Optional[Dict[str, float]] = None
    ) -> Tuple[float, float]:
        """
        Function for calculating the average flux for the
        ``filter_name``.

        Parameters
        ----------
        model_param : dict, None
            Model parameters. Should contain the 'scaling' value. Not
            used if set to ``None``.

        Returns
        -------
        Returns
        -------
        float
            Average flux (W m-2 um-1).
        float
            Uncertainty (W m-2 um-1).
        """

        specbox = self.get_spectrum(model_param=model_param, apply_mask=True)

        synphot = SyntheticPhotometry(self.filter_name)

        return synphot.spectrum_to_flux(
            specbox.wavelength, specbox.flux, error=specbox.flux
        )

    @typechecked
    def get_magnitude(
        self,
        model_param: Optional[Dict[str, float]] = None,
        distance: Optional[Tuple[float, float]] = None,
    ) -> Tuple[Tuple[float, Optional[float]], Tuple[Optional[float], Optional[float]]]:
        """
        Function for calculating the apparent magnitude for the
        ``filter_name``.

        Parameters
        ----------
        model_param : dict, None
            Model parameters. Should contain the 'scaling' value. Not
            used if set to ``None``.
        distance : tuple(float, float), None
            Distance and uncertainty to the calibration object (pc).
            Not used if set to ``None``, in which case the returned
            absolute magnitude is ``(None, None)``.

        Returns
        -------
        tuple(float, float)
            Apparent magnitude and uncertainty.
        tuple(float, float), tuple(None, None)
            Absolute magnitude and uncertainty.
        """

        specbox = self.get_spectrum(model_param=model_param, apply_mask=True)

        if np.count_nonzero(specbox.error) == 0:
            error = None
        else:
            error = specbox.error

        synphot = SyntheticPhotometry(self.filter_name)

        return synphot.spectrum_to_magnitude(
            specbox.wavelength, specbox.flux, error=error, distance=distance
        )



"""
Module with functionalities for calculating synthetic photometry.
"""

import os
import math
import warnings
import configparser

from typing import List, Optional, Union, Tuple

import h5py
import numpy as np

from typeguard import typechecked

from species.data.spec_data.spec_vega import add_vega
from species.read.read_filter import ReadFilter
from species.util.convert_util import apparent_to_absolute, parallax_to_distance


class SyntheticPhotometry:
    """
    Class for calculating synthetic photometry from a spectrum and also
    for converting between magnitudes and fluxes. Any filter from the
    `SVO Filter Profile Service <http://svo2.cab.inta-csic.es/svo/
    theory/fps/>`_ will be automatically downloaded and added to the
    database. Also the detector type (energy- or photon-counting) will
    be fetched. For a photon-counting detector, an additional
    wavelength factor is included in the integral for calculating the
    synthetic photometry, although typically the impact of the factor
    on the calculated flux is negligible. It is also important to note
    that by default the magnitude of Vega is set to 0.03 for all
    filters. The value can be adjusted in the `configuration file
    <https://species.readthedocs.io/en/latest/configuration.html>`_.
    """

    @typechecked
    def __init__(self, filter_name: str, zero_point: Optional[float] = None) -> None:
        """
        Parameters
        ----------
        filter_name : str
            Filter name by which the profile is stored in database.
            Any filter from the `SVO Filter Profile Service
            <http://svo2.cab.inta-csic.es/svo/theory/fps/>`_ will be
            automatically downloaded and added to the database.
        zero_point : float, None
            Zero-point flux (:math:`\\mathrm{W}`
            :math:`\\mathrm{m}^{-2}` :math:`\\mu\\mathrm{m}^{-1}`) for
            ``filter_name``. This flux is equalized to the magnitude of
            Vega, which by default is set to 0.03 for all filters. The
            value can be adjusted in the `configuration file <https://
            species.readthedocs.io/en/latest/configuration.html>`_.
            By default, the argument of ``zero_point`` is set to
            ``None``, in which case the zero point is calculated
            internally. The zero point can be accessed through
            ``zero_point`` attribute from instance of
            :class:`~species.phot.syn_phot.SyntheticPhotometry`.

        Returns
        -------
        NoneType
            None
        """

        self.filter_name = filter_name
        self.zero_point = zero_point
        self.filter_interp = None
        self.wavel_range = None

        config_file = os.path.join(os.getcwd(), "species_config.ini")

        config = configparser.ConfigParser()
        config.read(config_file)

        self.database = config["species"]["database"]
        self.data_folder = config["species"]["data_folder"]
        self.vega_mag = float(config["species"]["vega_mag"])

        read_filt = ReadFilter(self.filter_name)
        self.det_type = read_filt.detector_type()

        if self.zero_point is None:
            self.zero_point = self.calc_zero_point()

        else:
            warnings.warn(
                "Please note that a manually provided zero-point flux "
                "is by default equalized to a magnitude of 0.03 for "
                "all filters. The magnitude of Vega can be adjusted "
                "in the configuration file (see https://species."
                "readthedocs.io/en/latest/configuration.html) by "
                "setting the 'vega_mag' parameter. Currently the "
                f"parameter is set to {self.vega_mag}."
            )

    @typechecked
    def calc_zero_point(self) -> Union[float, np.float64]:
        """
        Internal function for calculating the zero point of the
        provided ``filter_name``. The zero point is here defined
        as the flux of Vega, which by default is set to a
        magnitude of 0.03 for all filters.

        Returns
        -------
        float
            Zero-point flux (:math:`\\mathrm{W}`
            :math:`\\mathrm{m}^{-2}` :math:`\\mu\\mathrm{m}^{-1}`).
        """

        if self.wavel_range is None:
            read_filt = ReadFilter(self.filter_name)
            self.wavel_range = read_filt.wavelength_range()

        with h5py.File(self.database, "r") as hdf5_file:
            if "spectra/calibration/vega" in hdf5_file:
                vega_found = True
            else:
                vega_found = False

        if not vega_found:
            with h5py.File(self.database, "a") as hdf5_file:
                add_vega(self.data_folder, hdf5_file)

        with h5py.File(self.database, "r") as hdf5_file:
            vega_spec = np.array(hdf5_file["spectra/calibration/vega"])

        wavelength = vega_spec[0,]
        flux = vega_spec[1,]

        wavelength_crop = wavelength[
            (wavelength > self.wavel_range[0]) & (wavelength < self.wavel_range[1])
        ]

        flux_crop = flux[
            (wavelength > self.wavel_range[0]) & (wavelength < self.wavel_range[1])
        ]

        return self.spectrum_to_flux(wavelength_crop, flux_crop)[0]

    @typechecked
    def spectrum_to_flux(
        self,
        wavelength: np.ndarray,
        flux: np.ndarray,
        error: Optional[np.ndarray] = None,
        threshold: Optional[float] = 0.01,
    ) -> Tuple[
        Union[float, np.float32, np.float64],
        Optional[Union[float, np.float32, np.float64]],
    ]:
        """
        Function for calculating the average flux from a spectrum and
        a filter profile. The uncertainty is propagated by sampling
        200 random values from the error distributions.

        Parameters
        ----------
        wavelength : np.ndarray
            Wavelength points (um).
        flux : np.ndarray
            Flux (:math:`\\mathrm{W}` :math:`\\mathrm{m}^{-2}`
            :math:`\\mu\\mathrm{m}^{-1}`).
        error : np.ndarray, None
            Uncertainty (:math:`\\mathrm{W}` :math:`\\mathrm{m}^{-2}`
            :math:`\\mu\\mathrm{m}^{-1}`). Not used if set to ``None``.
        threshold : float, None
            Transmission threshold (value between 0 and 1). If the
            minimum transmission value is larger than the threshold,
            a NaN is returned. This will happen if the input spectrum
            does not cover the full wavelength range of the filter
            profile. The parameter is not used if set to ``None``
            (default: 0.01).

        Returns
        -------
        float
            Average flux (:math:`\\mathrm{W}` :math:`\\mathrm{m}^{-2}`
            :math:`\\mu\\mathrm{m}^{-1}`).
        float, None
            Uncertainty (:math:`\\mathrm{W}` :math:`\\mathrm{m}^{-2}`
            :math:`\\mu\\mathrm{m}^{-1}`).
        """

        # Remove fluxes that are a NaN

        nan_idx = np.isnan(flux)

        if np.sum(nan_idx) > 0:
            warnings.warn(
                f"Found {np.sum(nan_idx)} fluxes with NaN. Removing "
                "these spectral fluxes from the input data before "
                "calculating synthetic photometry."
            )

            wavelength = wavelength[~nan_idx]
            flux = flux[~nan_idx]

            if error is not None:
                error = error[~nan_idx]

        if error is not None:
            # The error calculation requires the original
            # spectrum because spectrum_to_flux is used
            wavel_error = wavelength.copy()
            flux_error = flux.copy()

        if self.filter_interp is None:
            # Interpolate filter profile
            read_filt = ReadFilter(self.filter_name)
            self.filter_interp = read_filt.interpolate_filter()

            if self.wavel_range is None:
                # Set the wavel_range attribute to the
                # wavelength range of the filter profile
                self.wavel_range = read_filt.wavelength_range()

        if wavelength.size == 0:
            syn_flux = np.nan

            if error is not None:
                error_flux = np.nan
            else:
                error_flux = None

            indices = None

            warnings.warn(
                f"Calculation of the mean flux for {self.filter_name} "
                "is not possible because the wavelength array is "
                "empty. Returning a NaN for the flux."
            )

        else:
            indices = np.where(
                (self.wavel_range[0] <= wavelength)
                & (wavelength <= self.wavel_range[1])
            )[0]

        if indices is not None and indices.size < 2:
            syn_flux = np.nan

            if error is not None:
                error_flux = np.nan
            else:
                error_flux = None

            warnings.warn(
                "Calculating a synthetic flux requires more than "
                "one wavelength point. Photometry is set to NaN."
            )

        else:
            if threshold is None and (
                wavelength[0] > self.wavel_range[0]
                or wavelength[-1] < self.wavel_range[1]
            ):
                warnings.warn(
                    f"The filter profile of {self.filter_name} "
                    f"({self.wavel_range[0]:.4f}-"
                    f"{self.wavel_range[1]:.4f}) extends beyond "
                    f"the wavelength range of the spectrum "
                    f"({wavelength[0]:.4f}-{wavelength[-1]:.4f}). "
                    "The flux is set to NaN. Setting the 'threshold' "
                    "parameter will loosen the wavelength constraints."
                )

                syn_flux = np.nan

            else:
                wavelength = wavelength[indices]
                flux = flux[indices]

                transmission = self.filter_interp(wavelength)

                if (
                    threshold is not None
                    and (transmission[0] > threshold or transmission[-1] > threshold)
                    and (
                        wavelength[0] < self.wavel_range[0]
                        or wavelength[-1] > self.wavel_range[-1]
                    )
                ):
                    warnings.warn(
                        f"The filter profile of {self.filter_name} "
                        f"({self.wavel_range[0]:.4f}-{self.wavel_range[1]:.4f}) "
                        f"extends beyond the wavelength range of the spectrum "
                        f"({wavelength[0]:.4f}-{wavelength[-1]:.4f}). The flux "
                        f"is set to NaN. Increasing the 'threshold' parameter "
                        f"({threshold}) will loosen the wavelength constraint."
                    )

                    syn_flux = np.nan

                else:
                    indices = np.isnan(transmission)
                    indices = np.logical_not(indices)

                    if self.det_type == "energy":
                        # Energy counting detector
                        integrand1 = transmission[indices] * flux[indices]
                        integrand2 = transmission[indices]

                    elif self.det_type == "photon":
                        # Photon counting detector
                        integrand1 = (
                            wavelength[indices] * transmission[indices] * flux[indices]
                        )
                        integrand2 = wavelength[indices] * transmission[indices]

                    integral1 = np.trapz(integrand1, x=wavelength[indices])
                    integral2 = np.trapz(integrand2, x=wavelength[indices])

                    syn_flux = integral1 / integral2

        if error is not None and not np.any(np.isnan(error)) and not np.isnan(syn_flux):
            phot_random = np.zeros(200)

            for i in range(200):
                # Use the original spectrum size (i.e. wavel_error and flux_error)
                spec_random = (
                    flux_error
                    + np.random.normal(loc=0.0, scale=1.0, size=wavel_error.shape[0])
                    * error
                )

                phot_random[i] = self.spectrum_to_flux(
                    wavel_error, spec_random, error=None, threshold=threshold
                )[0]

            nan_idx = np.isnan(phot_random)

            if np.sum(nan_idx) > 0:
                warnings.warn(
                    f"{np.sum(nan_idx)} out of 200 samples "
                    "that are used for estimating the "
                    "uncertainty on the synthetic flux "
                    "are NaN so removing these samples."
                )

                phot_random = phot_random[~nan_idx]

            error_flux = np.std(phot_random)

        elif error is not None and np.any(np.isnan(error)):
            warnings.warn("Spectum contains NaN so can not calculate the error.")
            error_flux = None

        else:
            error_flux = None

        return syn_flux, error_flux

    @typechecked
    def spectrum_to_magnitude(
        self,
        wavelength: np.ndarray,
        flux: np.ndarray,
        error: Optional[Union[np.ndarray, List[np.ndarray]]] = None,
        parallax: Optional[Tuple[float, Optional[float]]] = None,
        distance: Optional[Tuple[float, Optional[float]]] = None,
        threshold: Optional[float] = 0.01,
    ) -> Tuple[
        Tuple[float, Optional[float]], Optional[Tuple[Optional[float], Optional[float]]]
    ]:
        """
        Function for calculating the apparent and absolute magnitude
        from a spectrum and a filter profile. The uncertainty is
        propagated by sampling 200 random values from the error
        distributions.

        Parameters
        ----------
        wavelength : np.ndarray
            Wavelength points (um).
        flux : np.ndarray
            Flux (:math:`\\mathrm{W}` :math:`\\mathrm{m}^{-2}`
            :math:`\\mu\\mathrm{m}^{-1}`).
        error : np.ndarray, list(np.ndarray), None
            Uncertainty (:math:`\\mathrm{W}` :math:`\\mathrm{m}^{-2}`
            :math:`\\mu\\mathrm{m}^{-1}`).
        parallax : tuple(float, float), None
            Parallax and uncertainty (mas). No absolute magnitude is
            calculated if set to ``None``. No error on the absolute
            magnitude is calculated if the ``error`` parameter is
            set to ``None``.
        distance : tuple(float, float), None
            Distance and uncertainty (pc). No absolute magnitude is
            calculated if set to ``None``. No error on the absolute
            magnitude is calculated if the ``error`` parameter is
            set to ``None``. This parameter is ignored if the
            ``parallax`` parameter is used.
        threshold : float, None
            Transmission threshold (value between 0 and 1). If the
            minimum transmission value is larger than the threshold,
            a NaN is returned. This will happen if the input spectrum
            does not cover the full wavelength range of the filter
            profile. The parameter is not used if set to ``None``
            (default: 0.01).

        Returns
        -------
        tuple(float, float)
            Apparent magnitude and uncertainty.
        tuple(float, float)
            Absolute magnitude and uncertainty.
        """

        # Remove fluxes that are a NaN

        nan_idx = np.isnan(flux)

        if np.sum(nan_idx) > 0:
            warnings.warn(
                f"Found {np.sum(nan_idx)} fluxes with NaN. Removing "
                "these spectral fluxes from the input data before "
                "calculating synthetic photometry."
            )

            wavelength = wavelength[~nan_idx]
            flux = flux[~nan_idx]

            if error is not None:
                error = error[~nan_idx]

        if parallax is not None:
            distance = parallax_to_distance(parallax)

        syn_flux = self.spectrum_to_flux(
            wavelength, flux, error=error, threshold=threshold
        )

        app_mag = self.vega_mag - 2.5 * math.log10(syn_flux[0] / self.zero_point)

        if error is not None and not np.any(np.isnan(error)):
            mag_random = np.zeros(200)

            for i in range(200):
                spec_random = (
                    flux
                    + np.random.normal(loc=0.0, scale=1.0, size=wavelength.shape[0])
                    * error
                )

                flux_random = self.spectrum_to_flux(
                    wavelength, spec_random, error=None, threshold=threshold
                )

                mag_random[i] = self.vega_mag - 2.5 * np.log10(
                    flux_random[0] / self.zero_point
                )

            nan_idx = np.isnan(mag_random)

            if np.sum(nan_idx) > 0:
                warnings.warn(
                    f"{np.sum(nan_idx)} out of 200 samples "
                    "that are used for estimating the "
                    "uncertainty on the synthetic magnitude "
                    "are NaN so removing these samples."
                )

                mag_random = mag_random[~nan_idx]

            error_app_mag = np.std(mag_random)

        elif error is not None and np.any(np.isnan(error)):
            warnings.warn("Spectum contains NaN so can not calculate the error.")
            error_app_mag = None

        else:
            error_app_mag = None

        if distance is None:
            abs_mag = None
            error_abs_mag = None

        else:
            abs_mag = app_mag - 5.0 * np.log10(distance[0]) + 5.0

            if error_app_mag is not None and distance[1] is not None:
                error_dist = distance[1] * (5.0 / (distance[0] * math.log(10.0)))
                error_abs_mag = math.sqrt(error_app_mag**2 + error_dist**2)

            else:
                error_abs_mag = None

        return (app_mag, error_app_mag), (abs_mag, error_abs_mag)

    @typechecked
    def magnitude_to_flux(
        self,
        magnitude: float,
        error: Optional[float] = None,
        zp_flux: Optional[float] = None,
    ) -> Tuple[
        Union[float, np.float32, np.float64],
        Optional[Union[float, np.float32, np.float64]],
    ]:
        """
        Function for converting a magnitude to a flux.

        Parameters
        ----------
        magnitude : float
            Magnitude.
        error : float, None
            Error on the magnitude. Not used if set to ``None``.
        zp_flux : float, None
            DEPRECATED: Zero-point flux (:math:`\\mathrm{W}`
            :math:`\\mathrm{m}^{-2}` :math:`\\mu\\mathrm{m}^{-1}`).
            This parameter is deprecated and will be removed in a
            future release. Please use the zero_point parameter
            of the constructor of
            :class:`~species.phot.syn_phot.SyntheticPhotometry`
            instead. By default, the zero point is calculated
            internally and stored as the ``zero_point`` attribute
            of an instance from
            :class:`~species.phot.syn_phot.SyntheticPhotometry`.

        Returns
        -------
        float
            Flux (:math:`\\mathrm{W}` :math:`\\mathrm{m}^{-2}`
            :math:`\\mu\\mathrm{m}^{-1}`).
        float, None
            Error (:math:`\\mathrm{W}` :math:`\\mathrm{m}^{-2}`
            :math:`\\mu\\mathrm{m}^{-1}`). The returned value is
            ``None`` if the argument of ``error`` is ``None``.
        """

        if zp_flux is None:
            flux = 10.0 ** (-0.4 * (magnitude - self.vega_mag)) * self.zero_point

        else:
            flux = 10.0 ** (-0.4 * (magnitude - self.vega_mag)) * zp_flux

            warnings.warn(
                "The 'zp_flux' parameter is deprecated "
                "and will be removed in a future release. "
                "Please use the 'zero_point' parameter "
                "of the SyntheticPhotometry constructor "
                "instead.",
                DeprecationWarning,
            )

        if error is None:
            error_flux = None

        else:
            error_upper = flux * (10.0 ** (0.4 * error) - 1.0)
            error_lower = flux * (1.0 - 10.0 ** (-0.4 * error))
            error_flux = (error_lower + error_upper) / 2.0

        return flux, error_flux

    @typechecked
    def flux_to_magnitude(
        self,
        flux: float,
        error: Optional[Union[float, np.ndarray]] = None,
        parallax: Optional[
            Union[
                Tuple[float, Optional[float]], Tuple[np.ndarray, Optional[np.ndarray]]
            ]
        ] = None,
        distance: Optional[
            Union[
                Tuple[float, Optional[float]], Tuple[np.ndarray, Optional[np.ndarray]]
            ]
        ] = None,
    ) -> Tuple[
        Union[Tuple[float, Optional[float]], Tuple[np.ndarray, Optional[np.ndarray]]],
        Union[
            Tuple[Optional[float], Optional[float]],
            Tuple[Optional[np.ndarray], Optional[np.ndarray]],
        ],
    ]:
        """
        Function for converting a flux into a magnitude.

        Parameters
        ----------
        flux : float, np.ndarray
            Flux (:math:`\\mathrm{W}` :math:`\\mathrm{m}^{-2}`
            :math:`\\mu\\mathrm{m}^{-1}`).
        error : float, np.ndarray, None
            Uncertainty (:math:`\\mathrm{W}` :math:`\\mathrm{m}^{-2}`
            :math:`\\mu\\mathrm{m}^{-1}`). Not used if set to None.
        parallax : tuple(float, float), , tuple(np.ndarray, np.ndarray), None
            Parallax and uncertainty (mas). The returned absolute
            magnitude is set to ``None`` in case ``parallax`` and
            ``distance`` are set to ``None``. The error is not
            propagated into the error on the absolute magnitude
            in case the parallax uncertainty is set to ``None``,
            for example ``parallax=(10., None)``.
        distance : tuple(float, float), tuple(np.ndarray, np.ndarray), None
            Distance and uncertainty (pc). The returned absolute
            magnitude is set to ``None`` in case ``distance`` and
            ``parallax`` are set to ``None``. The error is not
            propagated into the error on the absolute magnitude in
            case the distance uncertainty is set to ``None``, for
            example ``distance=(20., None)``. This parameter is
            ignored if the ``parallax`` parameter is used.

        Returns
        -------
        tuple(float, float), tuple(np.ndarray, np.ndarray)
            Apparent magnitude and uncertainty.
        tuple(float, float), tuple(np.ndarray, np.ndarray)
            Absolute magnitude and uncertainty.
        """

        if parallax is not None:
            distance = parallax_to_distance(parallax)

        if flux <= 0.0:
            raise ValueError(
                "Converting a flux into a magnitude "
                "is only possible if the argument of "
                "'flux' has a positive value."
            )

        app_mag = self.vega_mag - 2.5 * np.log10(flux / self.zero_point)

        if error is None:
            error_app_mag = None
            error_abs_mag = None

        else:
            if flux + error > 0.0:
                error_app_lower = app_mag - (
                    self.vega_mag - 2.5 * np.log10((flux + error) / self.zero_point)
                )

            else:
                error_app_lower = np.nan

            if flux - error > 0.0:
                error_app_upper = (
                    self.vega_mag - 2.5 * np.log10((flux - error) / self.zero_point)
                ) - app_mag

            else:
                error_app_upper = np.nan

            error_app_mag = np.nanmean([error_app_lower, error_app_upper])

            if np.isnan(error_app_mag):
                error_app_mag = None

                warnings.warn(
                    "This warning should not have occurred "
                    "since either error_app_lower and/or "
                    "error_app_upper should not be NaN."
                )

        if distance is None:
            abs_mag = None
            error_abs_mag = None

        else:
            abs_mag, error_abs_mag = apparent_to_absolute(
                (app_mag, error_app_mag), distance
            )

        return (app_mag, error_app_mag), (abs_mag, error_abs_mag)



"""
Utility functions for dust cross sections and extinction.
"""

import os
import configparser

from typing import Dict, List, Tuple, Union

import h5py
import numpy as np

from typeguard import typechecked
from scipy.interpolate import interp1d, RegularGridInterpolator
from scipy.stats import lognorm

from species.read.read_filter import ReadFilter
from species.data.misc_data.dust_data import add_cross_sections, add_optical_constants


@typechecked
def check_dust_database() -> str:
    """
    Function to check if the dust data is present in the
    database and add the data if needed.

    Returns
    -------
    str
        Path of the HDF5 database.
    """

    config_file = os.path.join(os.getcwd(), "species_config.ini")

    config = configparser.ConfigParser()
    config.read(config_file)

    database_path = config["species"]["database"]
    data_folder = config["species"]["data_folder"]

    with h5py.File(database_path, "a") as hdf5_file:
        if "dust" not in hdf5_file:
            add_optical_constants(data_folder, hdf5_file)
            add_cross_sections(data_folder, hdf5_file)

    return database_path


@typechecked
def log_normal_distribution(
    radius_g: float, sigma_g: float, n_bins: int
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    Function for returning a log-normal size distribution. See Eq. 9
    in Ackerman & Marley (2001).

    Parameters
    ----------
    radius_g : float
        Mean geometric radius (um).
    sigma_g : float
        Geometric standard deviation (dimensionless).
    n_bins : int
        Number of logarithmically-spaced radius bins.

    Returns
    -------
    np.ndarray
        Number of grains in each radius bin, normalized to a total of
        1 grain.
    np.ndarray
        Widths of the radius bins (um).
    np.ndarray
        Grain radii (um).
    """

    if sigma_g == 1.0:
        # The log-normal distribution is equal to a delta
        # function with sigma_g = 1
        radii = np.array([radius_g])
        r_width = np.array([np.nan])
        dn_grains = np.array([1.0])

    else:
        # Get the radius interval which contains 99.999%
        # of the distribution
        interval = lognorm.interval(
            1.0 - 1e-5, np.log(sigma_g), loc=0.0, scale=radius_g
        )

        # Create bin boundaries (um), so +1 because there
        # are n_bins+1 bin boundaries
        r_bins = np.logspace(np.log10(interval[0]), np.log10(interval[1]), n_bins + 1)

        # Width of the radius bins (um)
        r_width = np.diff(r_bins)

        # Grain radii (um) at which the size distribution is sampled
        radii = (r_bins[1:] + r_bins[:-1]) / 2.0

        # Number of grains per radius bin width, normalized to an
        # integrated value of 1 grain, that is,
        # np.sum(dn_dr*r_width) = 1
        # The log-normal distribution from Ackerman & Marley 2001
        # gives the same result as scipy.stats.lognorm.pdf with
        # s = log(sigma_g) and scale=radius_g
        dn_dr = lognorm.pdf(radii, s=np.log(sigma_g), loc=0.0, scale=radius_g)

        # Number of grains for each radius bin
        dn_grains = dn_dr * r_width

    return dn_grains, r_width, radii


@typechecked
def power_law_distribution(
    exponent: float, radius_min: float, radius_max: float, n_bins: int
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    Function for returning a power-law size distribution.

    Parameters
    ----------
    exponent : float
        Exponent of the power-law size distribution,
        dn/dr = r**exponent.
    radius_min : float
        Minimum grain radius (um).
    radius_max : float
        Maximum grain radius (um).
    n_bins : int
        Number of logarithmically-spaced radius bins.

    Returns
    -------
    np.ndarray
        Number of grains in each radius bin, normalized to a total
        of 1 grain.
    np.ndarray
        Widths of the radius bins (um).
    np.ndarray
        Grain radii (um).
    """

    # Create bin boundaries (um), so +1 because there
    # are n_sizes+1 bin boundaries (um)
    r_bins = np.logspace(np.log10(radius_min), np.log10(radius_max), n_bins + 1)

    # Width of the radius bins (um)
    r_width = np.diff(r_bins)

    # Grains radii (um) at which the size distribution is sampled
    radii = (r_bins[1:] + r_bins[:-1]) / 2.0

    # Number of grains per radius bins size
    dn_dr = radii**exponent

    # Normalize the size distribution to 1 grain
    dn_dr /= np.sum(r_width * dn_dr)

    # Number of grains for each radius bin
    dn_grains = dn_dr * r_width

    return dn_grains, r_width, radii


@typechecked
def dust_cross_section(
    dn_grains: np.ndarray,
    radii: np.ndarray,
    wavelength: float,
    n_index: float,
    k_index: float,
) -> np.float64:
    """
    Function for calculating the extinction cross section for a size
    distribution of dust grains.

    Parameters
    ----------
    dn_grains : np.ndarray
        Number of grains in each radius bin, normalized to a total
        of 1 grain.
    radii : np.ndarray
        Grain radii (um).
    wavelength : float
        Wavelength (um).
    n_index : float
        Real part of the refractive index.
    k_index : float
        Imaginary part of the refractive index.

    Returns
    -------
    float
        Extinction cross section (um2)
    """

    # Importing here because it causes an error with Python 3.10
    import PyMieScatt

    c_ext = 0.0

    for i, item in enumerate(radii):
        # From the PyMieScatt documentation: When using PyMieScatt,
        # pay close attention to the units of the your inputs and
        # outputs. Wavelength and particle diameters are always in
        # nanometers, efficiencies are unitless, cross-sections are
        # in nm2, coefficients are in Mm-1, and size distribution
        # concentration is always in cm-3.
        mie = PyMieScatt.MieQ(
            complex(n_index, k_index),
            wavelength * 1e3,  # (nm)
            2.0 * item * 1e3,  # diameter (nm)
            asDict=True,
            asCrossSection=False,
        )

        if "Qext" in mie:
            c_ext += np.pi * item**2 * mie["Qext"] * dn_grains[i]  # (um2)

        else:
            raise ValueError("Qext not found in the PyMieScatt dictionary.")

    return c_ext  # (um2)


@typechecked
def calc_reddening(
    filters_color: Tuple[str, str],
    extinction: Tuple[str, float],
    composition: str = "MgSiO3",
    structure: str = "crystalline",
    radius_g: float = 1.0,
) -> Tuple[float, float]:
    """
    Function for calculating the reddening of a color given the
    extinction for a given filter. A log-normal size distribution with
    a geometric standard deviation of 2 is used as parametrization for
    the grain sizes (Ackerman & Marley 2001).

    Parameters
    ----------
    filters_color : tuple(str, str)
        Filter names for which the extinction is calculated.
    extinction : str
        Filter name and extinction (mag).
    composition : str
        Dust composition ('MgSiO3' or 'Fe').
    structure : str
        Grain structure ('crystalline' or 'amorphous').
    radius_g : float
        Geometric radius of the grain size distribution (um).

    Returns
    -------
    float
        Extinction (mag) for ``filters_color[0]``.
    float
        Extinction (mag) for ``filters_color[1]``.
    """

    database_path = check_dust_database()

    h5_file = h5py.File(database_path, "r")

    filters = [extinction[0], filters_color[0], filters_color[1]]

    dn_grains, _, radii = log_normal_distribution(radius_g, 2.0, 100)

    c_ext = {}

    for item in filters:
        h5_file.close()

        read_filt = ReadFilter(item)
        filter_wavel = read_filt.mean_wavelength()

        h5_file = h5py.File(database_path, "r")

        if composition == "MgSiO3" and structure == "crystalline":
            for i in range(3):
                data = h5_file[f"dust/mgsio3/crystalline/axis_{i+1}"]

                wavel_index = (np.abs(data[:, 0] - filter_wavel)).argmin()

                # Average cross section of the three axes

                if i == 0:
                    c_ext[item] = (
                        dust_cross_section(
                            dn_grains,
                            radii,
                            data[wavel_index, 0],
                            data[wavel_index, 1],
                            data[wavel_index, 2],
                        )
                        / 3.0
                    )

                else:
                    c_ext[item] += (
                        dust_cross_section(
                            dn_grains,
                            radii,
                            data[wavel_index, 0],
                            data[wavel_index, 1],
                            data[wavel_index, 2],
                        )
                        / 3.0
                    )

        else:
            if composition == "MgSiO3" and structure == "amorphous":
                data = h5_file["dust/mgsio3/amorphous/"]

            elif composition == "Fe" and structure == "crystalline":
                data = h5_file["dust/fe/crystalline/"]

            elif composition == "Fe" and structure == "amorphous":
                data = h5_file["dust/fe/amorphous/"]

            wavel_index = (np.abs(data[:, 0] - filter_wavel)).argmin()

            c_ext[item] += (
                dust_cross_section(
                    dn_grains,
                    radii,
                    data[wavel_index, 0],
                    data[wavel_index, 1],
                    data[wavel_index, 2],
                )
                / 3.0
            )

    h5_file.close()

    return (
        extinction[1] * c_ext[filters_color[0]] / c_ext[extinction[0]],
        extinction[1] * c_ext[filters_color[1]] / c_ext[extinction[0]],
    )


@typechecked
def interp_lognorm(
    inc_phot: List[str],
    inc_spec: List[str],
) -> Tuple[
    Dict[str, Union[RegularGridInterpolator, List[RegularGridInterpolator]]],
    np.ndarray,
    np.ndarray,
]:
    """
    Function for interpolating the log-normal dust cross sections for
    each filter and spectrum.

    Parameters
    ----------
    inc_phot : list(str)
        List with filter names. Not used if the list is empty.
    inc_spec : list(str)
        List with the spectrum names (as stored in the database with
        :func:`~species.data.database.Database.add_object`). Not used
        if the list is empty.

    Returns
    -------
    dict
        Dictionary with the extinction cross section for each filter
        and spectrum.
    np.ndarray
        Grid points of the geometric mean radius.
    np.ndarray
        Grid points of the geometric standard deviation.
    """

    database_path = check_dust_database()

    with h5py.File(database_path, "r") as h5_file:
        cross_section = np.asarray(
            h5_file["dust/lognorm/mgsio3/crystalline/cross_section"]
        )
        wavelength = np.asarray(h5_file["dust/lognorm/mgsio3/crystalline/wavelength"])
        radius_g = np.asarray(h5_file["dust/lognorm/mgsio3/crystalline/radius_g"])
        sigma_g = np.asarray(h5_file["dust/lognorm/mgsio3/crystalline/sigma_g"])

    print("Grid boundaries of the dust opacities:")
    print(f"   - Wavelength (um) = {wavelength[0]:.2f} - {wavelength[-1]:.2f}")
    print(f"   - Geometric mean radius (um) = {radius_g[0]:.2e} - {radius_g[-1]:.2e}")
    print(f"   - Geometric standard deviation = {sigma_g[0]:.2f} - {sigma_g[-1]:.2f}")

    inc_phot.append("Generic/Bessell.V")

    cross_sections = {}

    for phot_item in inc_phot:
        read_filt = ReadFilter(phot_item)
        filt_trans = read_filt.get_filter()

        cross_phot = np.zeros((radius_g.shape[0], sigma_g.shape[0]))

        for i in range(radius_g.shape[0]):
            for j in range(sigma_g.shape[0]):
                cross_interp = interp1d(
                    wavelength, cross_section[:, i, j], kind="linear", bounds_error=True
                )

                cross_tmp = cross_interp(filt_trans[:, 0])

                integral1 = np.trapz(filt_trans[:, 1] * cross_tmp, x=filt_trans[:, 0])
                integral2 = np.trapz(filt_trans[:, 1], x=filt_trans[:, 0])

                # Filter-weighted average of the extinction cross section
                cross_phot[i, j] = integral1 / integral2

        cross_sections[phot_item] = RegularGridInterpolator(
            (radius_g, sigma_g), cross_phot, method="linear", bounds_error=True
        )

    print("Interpolating dust opacities...", end="")

    if len(inc_spec) > 0:
        cross_tmp = RegularGridInterpolator(
            (wavelength, radius_g, sigma_g),
            cross_section,
            method="linear",
            bounds_error=True,
        )

        cross_sections["spectrum"] = cross_tmp

    print(" [DONE]")

    return cross_sections, radius_g, sigma_g


@typechecked
def interp_powerlaw(
    inc_phot: List[str],
    inc_spec: List[str],
) -> Tuple[
    Dict[str, Union[RegularGridInterpolator, List[RegularGridInterpolator]]],
    np.ndarray,
    np.ndarray,
]:
    """
    Function for interpolating the power-law dust cross sections for
    each filter and spectrum.

    Parameters
    ----------
    inc_phot : list(str)
        List with filter names. Not used if the list is empty.
    inc_spec : list(str)
        List with the spectrum names (as stored in the database with
        :func:`~species.data.database.Database.add_object`). Not used
        if the list is empty.

    Returns
    -------
    dict
        Dictionary with the extinction cross section for each filter
        and spectrum.
    np.ndarray
        Grid points of the maximum radius.
    np.ndarray
        Grid points of the power-law exponent.
    """

    database_path = check_dust_database()

    with h5py.File(database_path, "r") as h5_file:
        cross_section = np.asarray(
            h5_file["dust/powerlaw/mgsio3/crystalline/cross_section"]
        )
        wavelength = np.asarray(h5_file["dust/powerlaw/mgsio3/crystalline/wavelength"])
        radius_max = np.asarray(h5_file["dust/powerlaw/mgsio3/crystalline/radius_max"])
        exponent = np.asarray(h5_file["dust/powerlaw/mgsio3/crystalline/exponent"])

    print("Grid boundaries of the dust opacities:")
    print(f"   - Wavelength (um) = {wavelength[0]:.2f} - {wavelength[-1]:.2f}")
    print(f"   - Maximum radius (um) = {radius_max[0]:.2e} - {radius_max[-1]:.2e}")
    print(f"   - Power-law exponent = {exponent[0]:.2f} - {exponent[-1]:.2f}")

    inc_phot.append("Generic/Bessell.V")

    cross_sections = {}

    for phot_item in inc_phot:
        read_filt = ReadFilter(phot_item)
        filt_trans = read_filt.get_filter()

        cross_phot = np.zeros((radius_max.shape[0], exponent.shape[0]))

        for i in range(radius_max.shape[0]):
            for j in range(exponent.shape[0]):
                cross_interp = interp1d(
                    wavelength, cross_section[:, i, j], kind="linear", bounds_error=True
                )

                cross_tmp = cross_interp(filt_trans[:, 0])

                integral1 = np.trapz(filt_trans[:, 1] * cross_tmp, x=filt_trans[:, 0])
                integral2 = np.trapz(filt_trans[:, 1], x=filt_trans[:, 0])

                # Filter-weighted average of the extinction cross section
                cross_phot[i, j] = integral1 / integral2

        cross_sections[phot_item] = RegularGridInterpolator(
            (radius_max, exponent), cross_phot, method="linear", bounds_error=True
        )

    print("Interpolating dust opacities...", end="")

    if len(inc_spec) > 0:
        cross_tmp = RegularGridInterpolator(
            (wavelength, radius_max, exponent),
            cross_section,
            method="linear",
            bounds_error=True,
        )

        cross_sections["spectrum"] = cross_tmp

    print(" [DONE]")

    return cross_sections, radius_max, exponent


@typechecked
def ism_extinction(
    av_mag: float, rv_red: float, wavelengths: Union[np.ndarray, List[float], float]
) -> np.ndarray:
    """
    Function for calculating the optical and IR extinction
    with the empirical relation from `Cardelli et al. (1989)
    <https://ui.adsabs.harvard.edu/abs/1989ApJ...345..245C/abstract>`_.

    Parameters
    ----------
    av_mag : float
        Extinction (mag) in the $V$ band.
    rv_red : float
        Reddening in the $V$ band, ``R_V = A_V / E(B-V)``.
    wavelengths : np.ndarray, list(float), float
        Array or list with the wavelengths (um) for which
        the extinction is calculated. It is also possible
        to provide a single value as float.

    Returns
    -------
    np.ndarray
        Extinction (mag) at ``wavelengths``.
    """

    if isinstance(wavelengths, float):
        wavelengths = np.array([wavelengths])

    elif isinstance(wavelengths, list):
        wavelengths = np.array(wavelengths)

    x_wavel = 1.0 / wavelengths
    y_wavel = x_wavel - 1.82

    a_coeff = np.zeros(x_wavel.size)
    b_coeff = np.zeros(x_wavel.size)

    indices = np.where(x_wavel < 1.1)[0]

    if len(indices) > 0:
        a_coeff[indices] = 0.574 * x_wavel[indices] ** 1.61
        b_coeff[indices] = -0.527 * x_wavel[indices] ** 1.61

    indices = np.where(x_wavel >= 1.1)[0]

    if len(indices) > 0:
        a_coeff[indices] = (
            1.0
            + 0.17699 * y_wavel[indices]
            - 0.50447 * y_wavel[indices] ** 2
            - 0.02427 * y_wavel[indices] ** 3
            + 0.72085 * y_wavel[indices] ** 4
            + 0.01979 * y_wavel[indices] ** 5
            - 0.77530 * y_wavel[indices] ** 6
            + 0.32999 * y_wavel[indices] ** 7
        )

        b_coeff[indices] = (
            1.41338 * y_wavel[indices]
            + 2.28305 * y_wavel[indices] ** 2
            + 1.07233 * y_wavel[indices] ** 3
            - 5.38434 * y_wavel[indices] ** 4
            - 0.62251 * y_wavel[indices] ** 5
            + 5.30260 * y_wavel[indices] ** 6
            - 2.09002 * y_wavel[indices] ** 7
        )

    return av_mag * (a_coeff + b_coeff / rv_red)


@typechecked
def apply_ism_ext(
    wavelengths: np.ndarray, flux: np.ndarray, v_band_ext: float, v_band_red: float
) -> np.ndarray:
    """
    Function for applying ISM extinction to a spectrum.

    wavelengths : np.ndarray
        Wavelengths (um) of the spectrum.
    flux : np.ndarray
        Fluxes (W m-2 um-1) of the spectrum.
    v_band_ext : float
        Extinction (mag) in the $V$ band.
    v_band_red : float
        Reddening in the $V$ band.

    Returns
    -------
    np.ndarray
        Fluxes (W m-2 um-1) with the extinction applied.
    """

    ext_mag = ism_extinction(v_band_ext, v_band_red, wavelengths)

    return flux * 10.0 ** (-0.4 * ext_mag)


@typechecked
def convert_to_av(
    filter_name: str, filter_ext: float, v_band_red: float = 3.1
) -> float:
    """
    Function for converting the extinction in any filter from
    the `SVO Filter Profile Service <http://svo2.cab.inta-csic.
    es/svo/theory/fps/>`_ to a visual extinction, :math:`A_V`.
    This is done by simply scaling the extinction so at the
    mean wavelength of the filter.

    filter_name : str
        Filter name for which the extinction will be
        converted to a visual extinction (i.e. :math:`A_V`).
    filter_ext : float
        Extinction (mag) for the ``filter_name``.
    v_band_red : float
        Reddening in the $V$ band.

    Returns
    -------
    float
        Visual extinction (i.e. :math:`A_V`) for which the
        extinction in the ``filter_name`` band is ``filter_ext``.
    """

    av_test = 1.0

    # Mean wavelength for filter_name
    read_filt = ReadFilter(filter_name)
    filt_wavel = np.array([read_filt.mean_wavelength()])

    # Calculate test extinction for A_V = 1.0
    # at mean wavelength of filter_name
    ext_ref = ism_extinction(av_test, v_band_red, filt_wavel)[0]

    # Scaling for A_V = 1.0 to the A_V for which
    # extinction of filter_name is filter_ext
    scaling = filter_ext / ext_ref

    # Should be the same as filter_ext
    # filter_ext_test = ism_extinction(scaling * av_test, v_band_red, filt_wavel)[0]

    return scaling * av_test



import os
import urllib.request

import numpy as np

from species.core import constants


def add_baraffe2015(database, input_path):
    """
    Function for adding the Baraffe et al. (2015)
    isochrone data to the database.

    Parameters
    ----------
    database : h5py._hl.files.File
        Database.
    input_path : str
        Folder where the data is located.

    Returns
    -------
    NoneType
        None
    """

    if not os.path.exists(input_path):
        os.makedirs(input_path)

    url_iso = (
        "http://perso.ens-lyon.fr/isabelle.baraffe/BHAC15dir/BHAC15_tracks+structure"
    )

    iso_tag = "Baraffe et al. (2015)"
    iso_size = "1.4 MB"
    db_tag = "baraffe2015"

    input_file = url_iso.rsplit("/", maxsplit=1)[-1]
    data_file = os.path.join(input_path, input_file)

    if not os.path.isfile(data_file):
        print(f"Downloading {iso_tag} isochrones ({iso_size})...", end="", flush=True)
        urllib.request.urlretrieve(url_iso, data_file)
        print(" [DONE]")

    # M/Ms, log t(yr), Teff, log(L/Ls), log(g), R/Rs,
    # Log(Li/Li0), log(Tc), log(ROc), Mrad, Rrad, k2conv, k2rad
    mass, log_age, teff, log_lum, log_g, radius, _, _, _, _, _, _, _ = np.loadtxt(
        data_file, unpack=True, skiprows=45, comments="!"
    )

    age = 1e-6 * 10.0**log_age  # (Myr)
    mass *= constants.M_SUN / constants.M_JUP  # (Msun) -> (Mjup)
    radius *= constants.R_SUN / constants.R_JUP  # (Msun) -> (Mjup)

    iso_data = np.column_stack([age, mass, teff, log_lum, log_g, radius])

    print(f"Adding isochrones: {iso_tag}...", end="", flush=True)

    dset = database.create_dataset(
        f"isochrones/{db_tag}/age", data=iso_data[:, 0]
    )  # (Myr)
    database.create_dataset(f"isochrones/{db_tag}/mass", data=iso_data[:, 1])  # (Mjup)
    database.create_dataset(f"isochrones/{db_tag}/teff", data=iso_data[:, 2])  # (K)
    database.create_dataset(
        f"isochrones/{db_tag}/log_lum", data=iso_data[:, 3]
    )  # log(L/Lsun)
    database.create_dataset(f"isochrones/{db_tag}/log_g", data=iso_data[:, 4])  # log(g)
    database.create_dataset(
        f"isochrones/{db_tag}/radius", data=iso_data[:, 5]
    )  # (Rjup)

    dset.attrs["model"] = db_tag

    print(" [DONE]")
    print(f"Database tag: {db_tag}")



"""
Module with reading functionalities for spectral libraries.
"""

import os

from configparser import ConfigParser
from typing import List, Optional

import h5py
import numpy as np

from typeguard import typechecked

from species.core.box import PhotometryBox, SpectrumBox, create_box
from species.data.spec_data.add_spec_data import add_spec_library
from species.phot.syn_phot import SyntheticPhotometry
from species.read.read_filter import ReadFilter


class ReadSpectrum:
    """
    Class for reading spectral library data from the database.
    """

    @typechecked
    def __init__(self, spec_library: str, filter_name: Optional[str] = None) -> None:
        """
        Parameters
        ----------
        spec_library : str
            Name of the spectral library ('irtf', 'spex', 'kesseli+2017',
            'bonnefoy+2014', 'allers+2013', or 'vega').
        filter_name : str, None
            Filter name for the wavelength range. Full spectra
            are read if the argument is set to ``None``.

        Returns
        -------
        NoneType
            None
        """

        self.spec_library = spec_library
        self.filter_name = filter_name

        if filter_name is None:
            self.wavel_range = None

        else:
            transmission = ReadFilter(filter_name)
            self.wavel_range = transmission.wavelength_range()

        config_file = os.path.join(os.getcwd(), "species_config.ini")

        config = ConfigParser()
        config.read(config_file)

        self.database = config["species"]["database"]
        self.data_folder = config["species"]["data_folder"]

    @typechecked
    def get_spectrum(
        self, sptypes: Optional[List[str]] = None, exclude_nan: bool = True
    ) -> SpectrumBox:
        """
        Function for selecting spectra from the database.

        Parameters
        ----------
        sptypes : list(str), None
            Spectral types to select from a library. The spectral types
            should be indicated with two characters (e.g. 'M5', 'L2',
            'T3'). All spectra are selected if set to ``None``. For
            each object in the ``spec_library``, the requested
            ``sptypes`` are first compared with the optical spectral
            type and, if not available, secondly the near-infrared
            spectral type.
        exclude_nan : bool
            Exclude wavelength points for which the flux is NaN.

        Returns
        -------
        species.core.box.SpectrumBox
            Box with the spectra.
        """

        with h5py.File(self.database, "a") as hdf5_file:
            if f"spectra/{self.spec_library}" not in hdf5_file:
                add_spec_library(
                    self.data_folder, hdf5_file, self.spec_library, sptypes
                )

        list_wavelength = []
        list_flux = []
        list_error = []
        list_name = []
        list_simbad = []
        list_sptype = []
        list_parallax = []
        list_spec_res = []

        with h5py.File(self.database, "r") as hdf5_file:
            for item in hdf5_file[f"spectra/{self.spec_library}"]:
                dset = hdf5_file[f"spectra/{self.spec_library}/{item}"]

                wavelength = dset[:, 0]  # (um)
                flux = dset[:, 1]  # (W m-2 um-1)
                error = dset[:, 2]  # (W m-2 um-1)

                if exclude_nan:
                    nan_index = np.isnan(flux)

                    wavelength = wavelength[~nan_index]
                    flux = flux[~nan_index]
                    error = error[~nan_index]

                if self.wavel_range is None:
                    wl_index = np.arange(0, len(wavelength), 1)

                else:
                    wl_index = (
                        (flux > 0.0)
                        & (wavelength > self.wavel_range[0])
                        & (wavelength < self.wavel_range[1])
                    )

                count = np.count_nonzero(wl_index)

                if count > 0:
                    index = np.where(wl_index)[0]

                    if index[0] > 0:
                        wl_index[index[0] - 1] = True

                    if index[-1] < len(wl_index) - 1:
                        wl_index[index[-1] + 1] = True

                    list_wavelength.append(wavelength[wl_index])
                    list_flux.append(flux[wl_index])
                    list_error.append(error[wl_index])

                    attrs = dset.attrs

                    if "name" in attrs:
                        if isinstance(dset.attrs["name"], str):
                            list_name.append(dset.attrs["name"])
                        else:
                            list_name.append(dset.attrs["name"].decode("utf-8"))
                    else:
                        list_name.append("")

                    if "simbad" in attrs:
                        if isinstance(dset.attrs["simbad"], str):
                            list_simbad.append(dset.attrs["simbad"])
                        else:
                            list_simbad.append(dset.attrs["simbad"].decode("utf-8"))
                    else:
                        list_simbad.append("")

                    if "sptype" in attrs:
                        if isinstance(dset.attrs["sptype"], str):
                            list_sptype.append(dset.attrs["sptype"])
                        else:
                            list_sptype.append(dset.attrs["sptype"].decode("utf-8"))
                    else:
                        list_sptype.append("None")

                    if "parallax" in attrs:
                        list_parallax.append(
                            (dset.attrs["parallax"], dset.attrs["parallax_error"])
                        )
                    else:
                        list_parallax.append((np.nan, np.nan))

                    if "spec_res" in attrs:
                        list_spec_res.append(dset.attrs["spec_res"])
                    else:
                        list_spec_res.append(np.nan)

                else:
                    list_wavelength.append(np.array([]))
                    list_flux.append(np.array([]))
                    list_error.append(np.array([]))
                    list_name.append("")
                    list_simbad.append("")
                    list_sptype.append("None")
                    list_parallax.append((np.nan, np.nan))
                    list_spec_res.append(np.nan)

        spec_box = SpectrumBox()
        spec_box.spec_library = self.spec_library

        if sptypes is not None:
            spec_box.wavelength = []
            spec_box.flux = []
            spec_box.error = []
            spec_box.name = []
            spec_box.simbad = []
            spec_box.sptype = []
            spec_box.parallax = []
            spec_box.spec_res = []

            for item in sptypes:
                for i, spec_item in enumerate(list_sptype):
                    if item == spec_item[:2]:
                        spec_box.wavelength.append(list_wavelength[i])
                        spec_box.flux.append(list_flux[i])
                        spec_box.error.append(list_error[i])
                        spec_box.name.append(list_name[i])
                        spec_box.simbad.append(list_simbad[i])
                        spec_box.sptype.append(list_sptype[i])
                        spec_box.parallax.append(list_parallax[i])
                        spec_box.spec_res.append(list_spec_res[i])

        else:
            spec_box.wavelength = list_wavelength
            spec_box.flux = list_flux
            spec_box.error = list_error
            spec_box.name = list_name
            spec_box.simbad = list_simbad
            spec_box.sptype = list_sptype
            spec_box.parallax = list_parallax
            spec_box.spec_res = list_spec_res

        return spec_box

    @typechecked
    def get_flux(self, sptypes: Optional[List[str]] = None) -> PhotometryBox:
        """
        Function for calculating the average flux density for the
        ``filter_name``.

        Parameters
        ----------
        sptypes : list(str), None
            Spectral types to select from a library. The spectral types
            should be indicated with two characters (e.g. 'M5', 'L2',
            'T3'). All spectra are selected if set to ``None``.

        Returns
        -------
        species.core.box.PhotometryBox
            Box with the synthetic photometry.
        """

        spec_box = self.get_spectrum(sptypes=sptypes, exclude_nan=True)

        n_spectra = len(spec_box.wavelength)

        filter_profile = ReadFilter(filter_name=self.filter_name)
        mean_wavel = filter_profile.mean_wavelength()

        wavelengths = np.full(n_spectra, mean_wavel)
        filters = np.full(n_spectra, self.filter_name)

        synphot = SyntheticPhotometry(filter_name=self.filter_name)

        phot_flux = []

        for i in range(n_spectra):
            flux = synphot.spectrum_to_flux(
                wavelength=spec_box.wavelength[i],
                flux=spec_box.flux[i],
                error=spec_box.error[i],
            )

            phot_flux.append(flux)

        phot_flux = np.asarray(phot_flux)

        return create_box(
            boxtype="photometry",
            name=spec_box.name,
            sptype=spec_box.sptype,
            wavelength=wavelengths,
            flux=phot_flux,
            app_mag=None,
            abs_mag=None,
            filter_name=filters,
        )

    @typechecked
    def get_magnitude(self, sptypes: Optional[List[str]] = None) -> PhotometryBox:
        """
        Function for calculating the apparent magnitude for the
        specified ``filter_name``.

        Parameters
        ----------
        sptypes : list(str), None
            Spectral types to select from the library. The spectral
            types should be indicated with two characters (e.g. 'M5',
            'L2', 'T3'). All spectra are selected if set to ``None``.

        Returns
        -------
        species.core.box.PhotometryBox
            Box with the synthetic photometry.
        """

        spec_box = self.get_spectrum(sptypes=sptypes, exclude_nan=True)

        n_spectra = len(spec_box.wavelength)

        filter_profile = ReadFilter(filter_name=self.filter_name)
        mean_wavel = filter_profile.mean_wavelength()

        wavelengths = np.full(n_spectra, mean_wavel)
        filters = np.full(n_spectra, self.filter_name)

        synphot = SyntheticPhotometry(filter_name=self.filter_name)

        app_mag = []
        abs_mag = []

        for i in range(n_spectra):
            if np.isnan(spec_box.parallax[i][0]):
                app_tmp = (np.nan, np.nan)
                abs_tmp = (np.nan, np.nan)

            else:
                app_tmp, abs_tmp = synphot.spectrum_to_magnitude(
                    spec_box.wavelength[i],
                    spec_box.flux[i],
                    error=spec_box.error[i],
                    parallax=(
                        float(spec_box.parallax[i][0]),
                        float(spec_box.parallax[i][1]),
                    ),
                )

            app_mag.append(app_tmp)
            abs_mag.append(abs_tmp)

        return create_box(
            boxtype="photometry",
            name=spec_box.name,
            sptype=spec_box.sptype,
            wavelength=wavelengths,
            flux=None,
            app_mag=np.asarray(app_mag),
            abs_mag=np.asarray(abs_mag),
            filter_name=filters,
        )



"""
Module with reading functionalities for atmospheric model spectra.
"""

import os
import warnings

from configparser import ConfigParser
from typing import Dict, List, Optional, Tuple, Union

import h5py
import numpy as np

from PyAstronomy.pyasl import rotBroad, fastRotBroad
from typeguard import typechecked
from scipy.integrate import simps
from scipy.interpolate import interp1d, RegularGridInterpolator

from species.core import constants
from species.core.box import (
    ColorColorBox,
    ColorMagBox,
    ModelBox,
    PhotometryBox,
    create_box,
)
from species.data.model_data.model_spectra import add_model_grid
from species.data.spec_data.spec_vega import add_vega
from species.phot.syn_phot import SyntheticPhotometry
from species.read.read_filter import ReadFilter
from species.read.read_planck import ReadPlanck
from species.util.convert_util import logg_to_mass
from species.util.dust_util import check_dust_database, ism_extinction, convert_to_av
from species.util.model_util import binary_to_single
from species.util.spec_util import smooth_spectrum


class ReadModel:
    """
    Class for reading a model spectrum from the database.
    """

    @typechecked
    def __init__(
        self,
        model: str,
        wavel_range: Optional[Tuple[float, float]] = None,
        filter_name: Optional[str] = None,
    ):
        """
        Parameters
        ----------
        model : str
            Name of the atmospheric model.
        wavel_range : tuple(float, float), None
            Wavelength range (um). Full spectrum is selected if set to
            ``None``. Not used if ``filter_name`` is not ``None``.
        filter_name : str, None
            Filter name that is used for the wavelength range. The
            ``wavel_range`` is used if set to ``None``.

        Returns
        -------
        NoneType
            None
        """

        self.model = model

        if self.model == "bt-settl":
            warnings.warn(
                "It is recommended to use the CIFIST "
                "grid of the BT-Settl, because it is "
                "a newer version. In that case, set "
                "model='bt-settl-cifist' when using "
                "add_model of Database."
            )

        self.spectrum_interp = None
        self.wl_points = None
        self.wl_index = None

        self.filter_name = filter_name
        self.wavel_range = wavel_range

        if self.filter_name is not None:
            transmission = ReadFilter(self.filter_name)

            self.wavel_range = transmission.wavelength_range()
            self.mean_wavelength = transmission.mean_wavelength()

        else:
            self.mean_wavelength = None

        config_file = os.path.join(os.getcwd(), "species_config.ini")

        config = ConfigParser()
        config.read(config_file)

        self.database = config["species"]["database"]
        self.data_folder = config["species"]["data_folder"]

        self.extra_param = [
            "radius",
            "distance",
            "parallax",
            "mass",
            "luminosity",
            "lognorm_radius",
            "lognorm_sigma",
            "lognorm_ext",
            "ism_ext",
            "ism_red",
            "powerlaw_max",
            "powerlaw_exp",
            "powerlaw_ext",
            "disk_teff",
            "disk_radius",
            "veil_a",
            "veil_b",
            "veil_ref",
            "vsini",
        ]

        # Test if the spectra are present in the database
        hdf5_file = self.open_database()
        hdf5_file.close()

    @typechecked
    def open_database(self) -> h5py._hl.files.File:
        """
        Internal function for opening the HDF5 database.

        Returns
        -------
        h5py._hl.files.File
            The HDF5 database.
        """

        with h5py.File(self.database, "r") as hdf5_file:
            if f"models/{self.model}" in hdf5_file:
                model_found = True

            else:
                model_found = False

                warnings.warn(
                    f"The '{self.model}' model spectra are not present "
                    "in the database. Will try to add the model grid. "
                    "If this does not work (e.g. currently without an "
                    "internet connection), then please use the "
                    "'add_model' method of 'Database' to add the "
                    "grid of spectra at a later moment."
                )

        if not model_found:
            # This will not work when using multiprocessing.
            # Model spectra should be added to the database
            # before running FitModel with MPI
            with h5py.File(self.database, "a") as hdf5_file:
                add_model_grid(self.model, self.data_folder, hdf5_file)

        return h5py.File(self.database, "r")

    @typechecked
    def wavelength_points(self) -> Tuple[np.ndarray, np.ndarray]:
        """
        Internal function for extracting the wavelength points and
        indices that are used.

        Returns
        -------
        np.ndarray
            Wavelength points (um).
        np.ndarray
            Array with the size of the original wavelength grid. The
            booleans indicate if a wavelength point was used.
        """

        with self.open_database() as hdf5_file:
            wl_points = np.array(hdf5_file[f"models/{self.model}/wavelength"])

        if self.wavel_range is None:
            wl_index = np.ones(wl_points.shape[0], dtype=bool)

        else:
            wl_index = (wl_points > self.wavel_range[0]) & (
                wl_points < self.wavel_range[1]
            )
            index = np.where(wl_index)[0]

            if index[0] - 1 >= 0:
                wl_index[index[0] - 1] = True

            if index[-1] + 1 < wl_index.size:
                wl_index[index[-1] + 1] = True

        return wl_points[wl_index], wl_index

    @typechecked
    def interpolate_model(self) -> None:
        """
        Internal function for linearly interpolating the
        full grid of model spectra.

        Returns
        -------
        NoneType
            None
        """

        grid_points = list(self.get_points().values())

        if self.wl_points is None:
            self.wl_points, self.wl_index = self.wavelength_points()

        with self.open_database() as hdf5_file:
            grid_flux = np.array(hdf5_file[f"models/{self.model}/flux"])
            grid_flux = grid_flux[..., self.wl_index]

        self.spectrum_interp = RegularGridInterpolator(
            grid_points,
            grid_flux,
            method="linear",
            bounds_error=False,
            fill_value=np.nan,
        )

    @typechecked
    def interpolate_grid(
        self,
        wavel_resample: Optional[np.ndarray] = None,
        spec_res: Optional[float] = None,
        **kwargs,
    ) -> None:
        """
        Internal function for linearly interpolating the grid of model
        spectra for a given filter or wavelength sampling.

        wavel_resample : np.ndarray, None
            Wavelength points for the resampling of the spectrum. The
            ``filter_name`` is used if set to ``None``.
        spec_res : float, None
            Spectral resolution that is used for the Gaussian filter.
            No smoothing is applied if the argument is set to ``None``.

        Returns
        -------
        NoneType
            None
        """

        self.interpolate_model()

        if "smooth" in kwargs:
            warnings.warn(
                "The 'smooth' parameter has been "
                "deprecated. Please set only the "
                "'spec_res' argument, which can be set "
                "to None for not applying a smoothing.",
                DeprecationWarning,
            )

            if not kwargs["smooth"] and spec_res is not None:
                spec_res = None

        points = []
        for item in self.get_points().values():
            points.append(list(item))

        param_list = self.get_parameters()
        n_param = len(param_list)

        dim_size = []
        for item in points:
            dim_size.append(len(item))

        if self.filter_name is not None:
            dim_size.append(1)
        else:
            dim_size.append(wavel_resample.size)

        flux_new = np.zeros(dim_size)

        if n_param == 1:
            model_param = {}

            for i, item_0 in enumerate(points[0]):
                model_param[param_list[0]] = item_0

                if self.filter_name is not None:
                    flux_new[i] = self.get_flux(model_param)[0]

                else:
                    flux_new[i, :] = self.get_model(
                        model_param,
                        spec_res=spec_res,
                        wavel_resample=wavel_resample,
                    ).flux

        elif n_param == 2:
            model_param = {}

            for i, item_0 in enumerate(points[0]):
                for j, item_1 in enumerate(points[1]):
                    model_param[param_list[0]] = item_0
                    model_param[param_list[1]] = item_1

                    if self.filter_name is not None:
                        flux_new[i, j] = self.get_flux(model_param)[0]

                    else:
                        flux_new[i, j, :] = self.get_model(
                            model_param,
                            spec_res=spec_res,
                            wavel_resample=wavel_resample,
                        ).flux

        elif n_param == 3:
            model_param = {}

            for i, item_0 in enumerate(points[0]):
                for j, item_1 in enumerate(points[1]):
                    for k, item_2 in enumerate(points[2]):
                        model_param[param_list[0]] = item_0
                        model_param[param_list[1]] = item_1
                        model_param[param_list[2]] = item_2

                        if self.filter_name is not None:
                            flux_new[i, j, k] = self.get_flux(model_param)[0]

                        else:
                            flux_new[i, j, k, :] = self.get_model(
                                model_param,
                                spec_res=spec_res,
                                wavel_resample=wavel_resample,
                            ).flux

        elif n_param == 4:
            model_param = {}

            for i, item_0 in enumerate(points[0]):
                for j, item_1 in enumerate(points[1]):
                    for k, item_2 in enumerate(points[2]):
                        for m, item_3 in enumerate(points[3]):
                            model_param[param_list[0]] = item_0
                            model_param[param_list[1]] = item_1
                            model_param[param_list[2]] = item_2
                            model_param[param_list[3]] = item_3

                            if self.filter_name is not None:
                                flux_new[i, j, k, m] = self.get_flux(model_param)[0]

                            else:
                                flux_new[i, j, k, m, :] = self.get_model(
                                    model_param,
                                    spec_res=spec_res,
                                    wavel_resample=wavel_resample,
                                ).flux

        elif n_param == 5:
            model_param = {}

            for i, item_0 in enumerate(points[0]):
                for j, item_1 in enumerate(points[1]):
                    for k, item_2 in enumerate(points[2]):
                        for m, item_3 in enumerate(points[3]):
                            for n, item_4 in enumerate(points[4]):
                                model_param[param_list[0]] = item_0
                                model_param[param_list[1]] = item_1
                                model_param[param_list[2]] = item_2
                                model_param[param_list[3]] = item_3
                                model_param[param_list[4]] = item_4

                                if self.filter_name is not None:
                                    flux_new[i, j, k, m, n] = self.get_flux(
                                        model_param
                                    )[0]

                                else:
                                    flux_new[i, j, k, m, n, :] = self.get_model(
                                        model_param,
                                        spec_res=spec_res,
                                        wavel_resample=wavel_resample,
                                    ).flux

        if self.filter_name is not None:
            transmission = ReadFilter(self.filter_name)
            self.wl_points = [transmission.mean_wavelength()]

        else:
            self.wl_points = wavel_resample

        self.spectrum_interp = RegularGridInterpolator(
            points,
            flux_new,
            method="linear",
            bounds_error=False,
            fill_value=np.nan,
        )

    @typechecked
    def apply_lognorm_ext(
        self,
        wavelength: np.ndarray,
        flux: np.ndarray,
        radius_interp: float,
        sigma_interp: float,
        v_band_ext: float,
    ) -> np.ndarray:
        """
        Internal function for applying extinction by dust to a
        spectrum.

        wavelength : np.ndarray
            Wavelengths (um) of the spectrum.
        flux : np.ndarray
            Fluxes (W m-2 um-1) of the spectrum.
        radius_interp : float
            Logarithm of the mean geometric radius (um) of the
            log-normal size distribution.
        sigma_interp : float
            Geometric standard deviation (dimensionless) of the
            log-normal size distribution.
        v_band_ext : float
            The extinction (mag) in the V band.

        Returns
        -------
        np.ndarray
            Fluxes (W m-2 um-1) with the extinction applied.
        """

        check_dust_database()

        with h5py.File(self.database, "r") as hdf5_file:
            # Read array with cross sections
            dust_cross = np.array(
                hdf5_file["dust/lognorm/mgsio3/crystalline/cross_section"]
            )

            # Read array with wavelengths
            dust_wavel = np.array(
                hdf5_file["dust/lognorm/mgsio3/crystalline/wavelength"]
            )

            # Read array with geometric radius
            dust_radius = np.array(
                hdf5_file["dust/lognorm/mgsio3/crystalline/radius_g"]
            )

            # Read array with geometric standard deviation
            dust_sigma = np.array(hdf5_file["dust/lognorm/mgsio3/crystalline/sigma_g"])

        if wavelength[0] < dust_wavel[0]:
            raise ValueError(
                f"The shortest wavelength ({wavelength[0]:.2e} um) for which the "
                f"spectrum will be calculated is smaller than the shortest "
                f"wavelength ({dust_wavel[0]:.2e} um) of the grid with dust cross "
                f"sections."
            )

        if wavelength[-1] > dust_wavel[-1]:
            raise ValueError(
                f"The longest wavelength ({wavelength[-1]:.2e} um) for which the "
                f"spectrum  will be calculated is larger than the longest wavelength "
                f"({dust_wavel[-1]:.2e} um) of the grid with dust cross sections."
            )

        # Interpolate cross sections as function of wavelength,
        # geometric radius, and geometric standard deviation

        dust_interp = RegularGridInterpolator(
            (dust_wavel, dust_radius, dust_sigma),
            dust_cross,
            method="linear",
            bounds_error=True,
        )

        # Read Bessell V-band filter profile

        read_filt = ReadFilter("Generic/Bessell.V")
        filt_trans = read_filt.get_filter()

        # Create empty array for V-band extinction

        cross_phot = np.zeros((dust_radius.shape[0], dust_sigma.shape[0]))

        for i in range(dust_radius.shape[0]):
            for j in range(dust_sigma.shape[0]):
                # Calculate the filter-weighted average of the extinction cross sections

                cross_interp = interp1d(
                    dust_wavel, dust_cross[:, i, j], kind="linear", bounds_error=True
                )

                cross_tmp = cross_interp(filt_trans[:, 0])

                integral1 = np.trapz(filt_trans[:, 1] * cross_tmp, x=filt_trans[:, 0])
                integral2 = np.trapz(filt_trans[:, 1], x=filt_trans[:, 0])

                cross_phot[i, j] = integral1 / integral2

        # Interpolate the grid with the V-band extinction as
        # function of geometric radius and standard deviation

        cross_interp = RegularGridInterpolator(
            (dust_radius, dust_sigma), cross_phot, method="linear", bounds_error=True
        )

        cross_v_band = cross_interp((10.0**radius_interp, sigma_interp))

        # Create arrays with fixed particle properties

        radius_full = np.full(wavelength.shape[0], 10.0**radius_interp)
        sigma_full = np.full(wavelength.shape[0], sigma_interp)

        # Interpolate the cross-sections for the input wavelengths
        # and the geometric radius and standard deviation

        cross_new = dust_interp(np.column_stack((wavelength, radius_full, sigma_full)))

        # Calculate number of grains for the requested V-band extinction

        n_grains = v_band_ext / cross_v_band / 2.5 / np.log10(np.exp(1.0))

        return flux * np.exp(-cross_new * n_grains)

    @typechecked
    def apply_powerlaw_ext(
        self,
        wavelength: np.ndarray,
        flux: np.ndarray,
        r_max_interp: float,
        exp_interp: float,
        v_band_ext: float,
    ) -> np.ndarray:
        """
        Internal function for applying extinction by dust to a
        spectrum.

        wavelength : np.ndarray
            Wavelengths (um) of the spectrum.
        flux : np.ndarray
            Fluxes (W m-2 um-1) of the spectrum.
        r_max_interp : float
            Maximum radius (um) of the power-law size distribution.
        exp_interp : float
            Exponent of the power-law size distribution.
        v_band_ext : float
            The extinction (mag) in the V band.

        Returns
        -------
        np.ndarray
            Fluxes (W m-2 um-1) with the extinction applied.
        """

        check_dust_database()

        with h5py.File(self.database, "r") as hdf5_file:
            # Read array with cross sections
            dust_cross = np.array(
                hdf5_file["dust/powerlaw/mgsio3/crystalline/cross_section"]
            )

            # Read array with wavelengths

            dust_wavel = np.array(
                hdf5_file["dust/powerlaw/mgsio3/crystalline/wavelength"]
            )

            # Read array with maximum particle radii

            dust_r_max = np.array(
                hdf5_file["dust/powerlaw/mgsio3/crystalline/radius_max"]
            )

            # Read array with power-law exponents

            dust_exp = np.array(hdf5_file["dust/powerlaw/mgsio3/crystalline/exponent"])

        # Interpolate cross sections as function of wavelength,
        # geometric radius, and geometric standard deviation

        dust_interp = RegularGridInterpolator(
            (dust_wavel, dust_r_max, dust_exp),
            dust_cross,
            method="linear",
            bounds_error=True,
        )

        if wavelength[0] < dust_wavel[0]:
            raise ValueError(
                f"The shortest wavelength ({wavelength[0]:.2e} um) for which the "
                f"spectrum will be calculated is smaller than the shortest "
                f"wavelength ({dust_wavel[0]:.2e} um) of the grid with dust cross "
                f"sections."
            )

        if wavelength[-1] > dust_wavel[-1]:
            raise ValueError(
                f"The longest wavelength ({wavelength[-1]:.2e} um) for which the "
                f"spectrum  will be calculated is larger than the longest wavelength "
                f"({dust_wavel[-1]:.2e} um) of the grid with dust cross sections."
            )

        # Read Bessell V-band filter profile

        read_filt = ReadFilter("Generic/Bessell.V")
        filt_trans = read_filt.get_filter()

        # Create empty array for V-band extinction

        cross_phot = np.zeros((dust_r_max.shape[0], dust_exp.shape[0]))

        for i in range(dust_r_max.shape[0]):
            for j in range(dust_exp.shape[0]):
                # Calculate the filter-weighted average of the extinction cross sections

                cross_interp = interp1d(
                    dust_wavel, dust_cross[:, i, j], kind="linear", bounds_error=True
                )

                cross_tmp = cross_interp(filt_trans[:, 0])

                integral1 = np.trapz(filt_trans[:, 1] * cross_tmp, x=filt_trans[:, 0])
                integral2 = np.trapz(filt_trans[:, 1], x=filt_trans[:, 0])

                cross_phot[i, j] = integral1 / integral2

        # Interpolate the grid with the V-band extinction as
        # function of geometric radius and standard deviation

        cross_interp = RegularGridInterpolator(
            (dust_r_max, dust_exp), cross_phot, method="linear", bounds_error=True
        )

        cross_v_band = cross_interp((10.0**r_max_interp, exp_interp))

        # Create arrays with fixed particle properties

        r_max_full = np.full(wavelength.shape[0], 10.0**r_max_interp)
        exp_full = np.full(wavelength.shape[0], exp_interp)

        # Interpolate the cross-sections for the input wavelengths
        # and the geometric radius and standard deviation

        cross_new = dust_interp(np.column_stack((wavelength, r_max_full, exp_full)))

        # Calculate number of grains for the requested V-band extinction

        n_grains = v_band_ext / cross_v_band / 2.5 / np.log10(np.exp(1.0))

        return flux * np.exp(-cross_new * n_grains)

    @staticmethod
    @typechecked
    def apply_ext_ism(
        wavelengths: np.ndarray, flux: np.ndarray, v_band_ext: float, v_band_red: float
    ) -> Tuple[np.ndarray, np.ndarray]:
        """
        Internal function for applying ISM extinction to a spectrum.

        wavelengths : np.ndarray
            Wavelengths (um) of the spectrum.
        flux : np.ndarray
            Fluxes (W m-2 um-1) of the spectrum.
        v_band_ext : float
            Extinction (mag) in the V band.
        v_band_red : float
            Reddening in the V band.

        Returns
        -------
        np.ndarray
            Fluxes (W m-2 um-1) with the extinction applied.
        np.ndarray
            Extinction (mag) as function of wavelength.
        """

        ext_mag = ism_extinction(v_band_ext, v_band_red, wavelengths)

        return flux * 10.0 ** (-0.4 * ext_mag), ext_mag

    @typechecked
    def get_model(
        self,
        model_param: Dict[str, float],
        spec_res: Optional[float] = None,
        wavel_resample: Optional[np.ndarray] = None,
        magnitude: bool = False,
        fast_rot_broad: bool = True,
        ext_filter: Optional[str] = None,
        **kwargs,
    ) -> ModelBox:
        """
        Function for extracting a model spectrum by linearly
        interpolating the model grid.

        Parameters
        ----------
        model_param : dict
            Dictionary with the model parameters and values. The values
            should be within the boundaries of the grid. The grid
            boundaries of the spectra in the database can be obtained
            with
            :func:`~species.read.read_model.ReadModel.get_bounds()`.
        spec_res : float, None
            Spectral resolution that is used for smoothing the spectrum
            with a Gaussian kernel. No smoothing is applied if the
            argument is set to ``None``.
        wavel_resample : np.ndarray, None
            Wavelength points (um) to which the spectrum is resampled.
            Optional smoothin with ``spec_res`` is applied for
            resampling with ``wavel_resample``. The wavelength points
            as stored in the database are used if the argument is set
            to ``None``.
        magnitude : bool
            Normalize the spectrum with a flux calibrated spectrum of
            Vega and return the magnitude instead of flux density.
        fast_rot_broad : bool
            Apply fast algorithm for the rotational broadening if set
            to ``True``, otherwise a slow but more accurate broadening
            is applied if set to ``False``. The fast algorithm will
            only provide an accurate broadening if the wavelength range
            of the spectrum is somewhat narrow (e.g. only the :math:`K`
            band). The argument is only used if the ``vsini`` parameter
            is included in the ``model_param`` dictionary.
        ext_filter : str, None
            Filter that is associated with the (optional) extinction
            parameter, ``ism_ext``. When the argument of ``ext_filter``
            is set to ``None``, the extinction is defined in the visual
            as usual (i.e. :math:`A_V`). By providing a filter name
            from the `SVO Filter Profile Service <http://svo2.cab.
            inta-csic.es/svo/theory/fps/>`_ as argument then the
            extinction ``ism_ext`` is defined in that filter instead
            of the $V$ band.

        Returns
        -------
        species.core.box.ModelBox
            Box with the model spectrum.
        """

        if "smooth" in kwargs:
            warnings.warn(
                "The 'smooth' parameter has been "
                "deprecated. Please set only the "
                "'spec_res' argument, which can be set "
                "to None for not applying a smoothing.",
                DeprecationWarning,
            )

            if not kwargs["smooth"] and spec_res is not None:
                spec_res = None

        # Get grid boundaries

        grid_bounds = self.get_bounds()

        # Check if all parameters are present and within the grid boundaries

        param_list = self.get_parameters()

        for key in param_list:
            if key not in model_param.keys():
                raise ValueError(
                    f"The '{key}' parameter is required by '{self.model}'. "
                    f"The mandatory parameters are {param_list}."
                )

            if model_param[key] < grid_bounds[key][0]:
                raise ValueError(
                    f"The input value of '{key}' is smaller than the lower "
                    f"boundary of the model grid ({model_param[key]} < "
                    f"{grid_bounds[key][0]})."
                )

            if model_param[key] > grid_bounds[key][1]:
                raise ValueError(
                    f"The input value of '{key}' is larger than the upper "
                    f"boundary of the model grid ({model_param[key]} > "
                    f"{grid_bounds[key][1]})."
                )

        # Print a warning if redundant parameters are included in the dictionary

        ignore_param = []

        param_list = self.get_parameters()

        for key in model_param.keys():
            if (
                key not in param_list
                and key not in self.extra_param
                and not key.startswith("phot_ext_")
            ):
                warnings.warn(
                    f"The '{key}' parameter is not required by "
                    f"'{self.model}' so the parameter will be "
                    f"ignored. The mandatory parameters are "
                    f"{param_list}."
                )

                ignore_param.append(key)

        # Interpolate the model grid

        if self.spectrum_interp is None:
            self.interpolate_model()

        # Set the wavelength range

        if self.wavel_range is None:
            wl_points = self.get_wavelengths()
            self.wavel_range = (wl_points[0], wl_points[-1])

        # Create a list with the parameter values

        check_param = [
            "teff",
            "logg",
            "feh",
            "c_o_ratio",
            "fsed",
            "log_kzz",
            "ad_index",
        ]

        parameters = []
        for item in check_param:
            if item in model_param and item not in ignore_param:
                parameters.append(model_param[item])

        # Check if the ext_filter should be adjusted
        # to the name that is extracted from the
        # phot_ext_{ext_filter} parameter

        if ext_filter is None:
            for param_item in model_param:
                if param_item.startswith("phot_ext_"):
                    ext_filter = param_item[9:]

        # Interpolate the spectrum from the grid

        flux = self.spectrum_interp(parameters)[0]

        # Add the radius to the parameter dictionary if the mass if given

        if "mass" in model_param and "radius" not in model_param:
            mass = 1e3 * model_param["mass"] * constants.M_JUP  # (g)
            radius = np.sqrt(
                1e3 * constants.GRAVITY * mass / (10.0 ** model_param["logg"])
            )  # (cm)
            model_param["radius"] = 1e-2 * radius / constants.R_JUP  # (Rjup)

        # Apply (radius/distance)^2 scaling

        if "radius" in model_param and "parallax" in model_param:
            scaling = (model_param["radius"] * constants.R_JUP) ** 2 / (
                1e3 * constants.PARSEC / model_param["parallax"]
            ) ** 2

            flux *= scaling

        elif "radius" in model_param and "distance" in model_param:
            scaling = (model_param["radius"] * constants.R_JUP) ** 2 / (
                model_param["distance"] * constants.PARSEC
            ) ** 2

            flux *= scaling

        elif "flux_scaling" in model_param:
            flux *= model_param["flux_scaling"]

        elif "log_flux_scaling" in model_param:
            flux *= 10.0 ** model_param["log_flux_scaling"]

        # Add optional offset to the flux

        if "flux_offset" in model_param:
            flux += model_param["flux_offset"]

        # Add blackbody disk component to the spectrum

        if "disk_teff" in model_param and "disk_radius" in model_param:
            disk_param = {
                "teff": model_param["disk_teff"],
                "radius": model_param["disk_radius"],
            }

            if "parallax" in model_param:
                disk_param["parallax"] = model_param["parallax"]

            elif "distance" in model_param:
                disk_param["distance"] = model_param["distance"]

            readplanck = ReadPlanck(
                (0.9 * self.wavel_range[0], 1.1 * self.wavel_range[-1])
            )

            planck_box = readplanck.get_spectrum(disk_param, spec_res=spec_res)

            flux_interp = interp1d(
                planck_box.wavelength, planck_box.flux, bounds_error=False
            )
            flux += flux_interp(self.wl_points)

            # flux += spectres.spectres(
            #     self.wl_points, planck_box.wavelength, planck_box.flux
            # )

        # Create ModelBox with the spectrum

        model_box = create_box(
            boxtype="model",
            model=self.model,
            wavelength=self.wl_points,
            flux=flux,
            parameters=model_param,
            quantity="flux",
            spec_res=spec_res,
        )

        # Apply rotational broadening vsin(i) in km/s

        if "vsini" in model_param:
            spec_interp = interp1d(
                model_box.wavelength, model_box.flux, bounds_error=False
            )

            wavel_new = np.linspace(
                model_box.wavelength[0],
                model_box.wavelength[-1],
                2 * model_box.wavelength.size,
            )

            if fast_rot_broad:
                flux_broad = fastRotBroad(
                    wvl=wavel_new,
                    flux=spec_interp(wavel_new),
                    epsilon=0.0,
                    vsini=model_param["vsini"],
                    effWvl=None,
                )

            else:
                flux_broad = rotBroad(
                    wvl=wavel_new,
                    flux=spec_interp(wavel_new),
                    epsilon=0.0,
                    vsini=model_param["vsini"],
                    edgeHandling="firstlast",
                )

            spec_interp = interp1d(wavel_new, flux_broad, bounds_error=False)
            model_box.flux = spec_interp(model_box.wavelength)

        # Apply veiling

        if (
            "veil_a" in model_param
            and "veil_b" in model_param
            and "veil_ref" in model_param
        ):
            lambda_ref = 0.5  # (um)

            veil_flux = model_param["veil_ref"] + model_param["veil_b"] * (
                model_box.wavelength - lambda_ref
            )

            model_box.flux = model_param["veil_a"] * model_box.flux + veil_flux

        # Apply extinction

        if (
            "lognorm_radius" in model_param
            and "lognorm_sigma" in model_param
            and "lognorm_ext" in model_param
        ):
            model_box.flux = self.apply_lognorm_ext(
                model_box.wavelength,
                model_box.flux,
                model_param["lognorm_radius"],
                model_param["lognorm_sigma"],
                model_param["lognorm_ext"],
            )

        if (
            "powerlaw_max" in model_param
            and "powerlaw_exp" in model_param
            and "powerlaw_ext" in model_param
        ):
            model_box.flux = self.apply_powerlaw_ext(
                model_box.wavelength,
                model_box.flux,
                model_param["powerlaw_max"],
                model_param["powerlaw_exp"],
                model_param["powerlaw_ext"],
            )

        if "ism_ext" in model_param or ext_filter is not None:
            ism_reddening = model_param.get("ism_red", 3.1)

            if ext_filter is not None:
                ism_ext_av = convert_to_av(
                    filter_name=ext_filter,
                    filter_ext=model_param[f"phot_ext_{ext_filter}"],
                    v_band_red=ism_reddening,
                )

            else:
                ism_ext_av = model_param["ism_ext"]

            model_box.flux, ext_mag = self.apply_ext_ism(
                model_box.wavelength,
                model_box.flux,
                ism_ext_av,
                ism_reddening,
            )

            idx_select = ext_mag >= 0.0
            model_box.wavelength = model_box.wavelength[idx_select]
            model_box.flux = model_box.flux[idx_select]

        # Smooth the spectrum

        if spec_res is not None:
            model_box.flux = smooth_spectrum(
                model_box.wavelength, model_box.flux, spec_res
            )

        # Resample the spectrum

        if wavel_resample is not None:
            flux_interp = interp1d(
                model_box.wavelength, model_box.flux, bounds_error=False
            )
            model_box.flux = flux_interp(wavel_resample)

            # model_box.flux = spectres.spectres(
            #     wavel_resample,
            #     model_box.wavelength,
            #     model_box.flux,
            #     spec_errs=None,
            #     fill=np.nan,
            #     verbose=True,
            # )

            model_box.wavelength = wavel_resample

        # elif spec_res is not None and not smooth:
        #     index = np.where(np.isnan(model_box.flux))[0]
        #
        #     if index.size > 0:
        #         raise ValueError(
        #             "Flux values should not contains NaNs. Please make sure that "
        #             "the parameter values and the wavelength range are within "
        #             "the grid boundaries as stored in the database."
        #         )
        #
        #     wavel_resample = create_wavelengths(
        #         (self.wl_points[0], self.wl_points[-1]), spec_res
        #     )
        #
        #     indices = np.where(
        #         (wavel_resample > self.wl_points[0])
        #         & (wavel_resample < self.wl_points[-2])
        #     )[0]
        #
        #     wavel_resample = wavel_resample[indices]
        #
        #     flux_interp = interp1d(
        #         model_box.wavelength, model_box.flux, bounds_error=False
        #     )
        #     model_box.flux = flux_interp(wavel_resample)
        #
        #     # model_box.flux = spectres.spectres(
        #     #     wavel_resample,
        #     #     model_box.wavelength,
        #     #     model_box.flux,
        #     #     spec_errs=None,
        #     #     fill=np.nan,
        #     #     verbose=True,
        #     # )
        #
        #     model_box.wavelength = wavel_resample

        # Convert flux to magnitude

        if magnitude:
            with h5py.File(self.database, "a") as hdf5_file:
                if "spectra/calibration/vega" not in hdf5_file:
                    add_vega(self.data_folder, hdf5_file)

                vega_spec = np.array(hdf5_file["spectra/calibration/vega"])

            flux_interp = interp1d(vega_spec[0,], vega_spec[1,])

            flux_vega = flux_interp(model_box.wavelength)

            # flux_vega, _ = spectres.spectres(
            #     model_box.wavelength,
            #     calib_box.wavelength,
            #     calib_box.flux,
            #     spec_errs=calib_box.error,
            #     fill=np.nan,
            #     verbose=True,
            # )

            model_box.flux = -2.5 * np.log10(model_box.flux / flux_vega)
            model_box.quantity = "magnitude"

        # Check if the contains NaNs

        if np.isnan(np.sum(model_box.flux)):
            warnings.warn(
                "The resampled spectrum contains "
                f"{np.sum(np.isnan(model_box.flux))} "
                "NaNs, probably because the original "
                "wavelength range does not fully encompass "
                "the new wavelength range. This happened with "
                f"the following parameters: {model_param}."
            )

        # Add the luminosity to the parameter dictionary

        if "radius" in model_box.parameters:
            model_box.parameters["luminosity"] = (
                4.0
                * np.pi
                * (model_box.parameters["radius"] * constants.R_JUP) ** 2
                * constants.SIGMA_SB
                * model_box.parameters["teff"] ** 4.0
                / constants.L_SUN
            )  # (Lsun)

        # Add the blackbody disk component to the luminosity

        if (
            "disk_teff" in model_box.parameters
            and "disk_radius" in model_box.parameters
        ):
            model_box.parameters["luminosity"] += (
                4.0
                * np.pi
                * (model_box.parameters["disk_radius"] * constants.R_JUP) ** 2
                * constants.SIGMA_SB
                * model_box.parameters["disk_teff"] ** 4.0
                / constants.L_SUN
            )  # (Lsun)

        # Add the planet mass to the parameter dictionary

        if "radius" in model_param and "logg" in model_param:
            model_param["mass"] = logg_to_mass(
                model_param["logg"], model_param["radius"]
            )

        return model_box

    @typechecked
    def get_data(
        self,
        model_param: Dict[str, float],
        spec_res: Optional[float] = None,
        wavel_resample: Optional[np.ndarray] = None,
        ext_filter: Optional[str] = None,
    ) -> ModelBox:
        """
        Function for selecting a model spectrum (without interpolation)
        for a set of parameter values that coincide with the grid
        points. The stored grid points can be inspected with
        :func:`~species.read.read_model.ReadModel.get_points`.

        Parameters
        ----------
        model_param : dict
            Model parameters and values. Only discrete values from the
            original grid are possible. Else, the nearest grid values
            are selected.
        spec_res : float, None
            Spectral resolution that is used for smoothing the spectrum
            with a Gaussian kernel. No smoothing is applied to the
            spectrum if the argument is set to ``None``.
        wavel_resample : np.ndarray, None
            Wavelength points (um) to which the spectrum will be
            resampled. In that case, ``spec_res`` can still be used for
            smoothing the spectrum with a Gaussian kernel. The original
            wavelength points are used if the argument is set to
            ``None``.
        ext_filter : str, None
            Filter that is associated with the (optional) extinction
            parameter, ``ism_ext``. When the argument of ``ext_filter``
            is set to ``None``, the extinction is defined in the visual
            as usual (i.e. :math:`A_V`). By providing a filter name
            from the `SVO Filter Profile Service <http://svo2.cab.
            inta-csic.es/svo/theory/fps/>`_ as argument then the
            extinction ``ism_ext`` is defined in that filter instead
            of the $V$ band.

        Returns
        -------
        species.core.box.ModelBox
            Box with the model spectrum.
        """

        # Check if all parameters are present

        param_list = self.get_parameters()

        for key in param_list:
            if key not in model_param.keys():
                raise ValueError(
                    f"The '{key}' parameter is required by '{self.model}'. "
                    f"The mandatory parameters are {param_list}."
                )

        # Print a warning if redundant parameters are included in the dictionary

        ignore_param = []

        for key in model_param.keys():
            if key not in param_list and key not in self.extra_param:
                warnings.warn(
                    f"The '{key}' parameter is not required by "
                    f"'{self.model}' so the parameter will be "
                    f"ignored. The mandatory parameters are "
                    f"{param_list}."
                )

                ignore_param.append(key)

        # Set the wavelength range

        if self.wavel_range is None:
            wl_points = self.get_wavelengths()
            self.wavel_range = (wl_points[0], wl_points[-1])

        # Model parameters to check

        check_param = [
            "teff",
            "logg",
            "feh",
            "c_o_ratio",
            "fsed",
            "log_kzz",
            "ad_index",
        ]

        # Create lists with the parameter names and values

        param_key = []
        param_val = []

        for param_item in check_param:
            if param_item in model_param and param_item not in ignore_param:
                param_key.append(param_item)
                param_val.append(model_param[param_item])

        # Check if the ext_filter should be adjusted
        # to the name that is extracted from the
        # phot_ext_{ext_filter} parameter

        if ext_filter is None:
            for param_item in model_param:
                if param_item.startswith("phot_ext_"):
                    ext_filter = param_item[9:]

        # Read the wavelength grid and the indices that will
        # be used for the wavelength range

        wl_points, wl_index = self.wavelength_points()

        # Open de HDF5 database

        with self.open_database() as hdf5_file:
            # Read the grid of fluxes from the database

            flux = np.array(hdf5_file[f"models/{self.model}/flux"])

            # Find the indices of the grid points for which the spectrum will be extracted

            indices = []

            for i, item in enumerate(param_key):
                data = np.array(hdf5_file[f"models/{self.model}/{item}"])
                data_index = np.argwhere(
                    np.round(data, 4) == np.round(model_param[item], 4)
                )

                if len(data_index) == 0:
                    raise ValueError(
                        f"The parameter {item}={param_val[i]} is not found."
                    )

                data_index = data_index[0]

                indices.append(data_index[0])

        # Append the wavelength indices to the list of grid indices

        indices.append(wl_index)

        # Extract the spectrum at the requested grid point and
        # wavelength range

        flux = flux[tuple(indices)]

        # Apply (radius/distance)^2 scaling

        if "radius" in model_param and "parallax" in model_param:
            scaling = (model_param["radius"] * constants.R_JUP) ** 2 / (
                1e3 * constants.PARSEC / model_param["parallax"]
            ) ** 2

            flux *= scaling

        elif "radius" in model_param and "distance" in model_param:
            scaling = (model_param["radius"] * constants.R_JUP) ** 2 / (
                model_param["distance"] * constants.PARSEC
            ) ** 2

            flux *= scaling

        elif "flux_scaling" in model_param:
            flux *= model_param["flux_scaling"]

        elif "log_flux_scaling" in model_param:
            flux *= 10.0 ** model_param["log_flux_scaling"]

        # Add optional offset to the flux

        if "flux_offset" in model_param:
            flux += model_param["flux_offset"]

        # Add blackbody disk component to the spectrum

        if "disk_teff" in model_param and "disk_radius" in model_param:
            disk_param = {
                "teff": model_param["disk_teff"],
                "radius": model_param["disk_radius"],
            }

            if "parallax" in model_param:
                disk_param["parallax"] = model_param["parallax"]

            elif "distance" in model_param:
                disk_param["distance"] = model_param["distance"]

            readplanck = ReadPlanck(
                (0.9 * self.wavel_range[0], 1.1 * self.wavel_range[-1])
            )

            planck_box = readplanck.get_spectrum(disk_param, spec_res=spec_res)

            flux_interp = interp1d(
                planck_box.wavelength, planck_box.flux, bounds_error=False
            )
            flux += flux_interp(wl_points)

            # flux += spectres.spectres(wl_points, planck_box.wavelength, planck_box.flux)

        # Create ModelBox with the spectrum

        model_box = create_box(
            boxtype="model",
            model=self.model,
            wavelength=wl_points,
            flux=flux,
            parameters=model_param,
            quantity="flux",
            spec_res=spec_res,
        )

        # Apply extinction

        if (
            "lognorm_radius" in model_param
            and "lognorm_sigma" in model_param
            and "lognorm_ext" in model_param
        ):
            model_box.flux = self.apply_lognorm_ext(
                model_box.wavelength,
                model_box.flux,
                model_param["lognorm_radius"],
                model_param["lognorm_sigma"],
                model_param["lognorm_ext"],
            )

        if (
            "powerlaw_max" in model_param
            and "powerlaw_exp" in model_param
            and "powerlaw_ext" in model_param
        ):
            model_box.flux = self.apply_powerlaw_ext(
                model_box.wavelength,
                model_box.flux,
                model_param["powerlaw_max"],
                model_param["powerlaw_exp"],
                model_param["powerlaw_ext"],
            )

        if "ism_ext" in model_param or ext_filter is not None:
            ism_reddening = model_param.get("ism_red", 3.1)

            if ext_filter is not None:
                ism_ext_av = convert_to_av(
                    filter_name=ext_filter,
                    filter_ext=model_param[f"phot_ext_{ext_filter}"],
                    v_band_red=ism_reddening,
                )

            else:
                ism_ext_av = model_param["ism_ext"]

            model_box.flux, ext_mag = self.apply_ext_ism(
                model_box.wavelength,
                model_box.flux,
                ism_ext_av,
                ism_reddening,
            )

            idx_select = ext_mag >= 0.0
            model_box.wavelength = model_box.wavelength[idx_select]
            model_box.flux = model_box.flux[idx_select]

        # Smooth the spectrum

        if spec_res is not None:
            model_box.flux = smooth_spectrum(
                model_box.wavelength, model_box.flux, spec_res
            )

        # Resample the spectrum

        if wavel_resample is not None:
            flux_interp = interp1d(
                model_box.wavelength, model_box.flux, bounds_error=False
            )
            model_box.flux = flux_interp(wavel_resample)

            # model_box.flux = spectres.spectres(
            #     wavel_resample,
            #     model_box.wavelength,
            #     model_box.flux,
            #     spec_errs=None,
            #     fill=np.nan,
            #     verbose=True,
            # )

            model_box.wavelength = wavel_resample

        # Add the planet mass to the parameter dictionary

        if "radius" in model_param and "logg" in model_param:
            model_param["mass"] = logg_to_mass(
                model_param["logg"], model_param["radius"]
            )

        # Add the luminosity to the parameter dictionary

        if "radius" in model_box.parameters:
            model_box.parameters["luminosity"] = (
                4.0
                * np.pi
                * (model_box.parameters["radius"] * constants.R_JUP) ** 2
                * constants.SIGMA_SB
                * model_box.parameters["teff"] ** 4.0
                / constants.L_SUN
            )  # (Lsun)

        # Add the blackbody disk component to the luminosity

        if (
            "disk_teff" in model_box.parameters
            and "disk_radius" in model_box.parameters
        ):
            model_box.parameters["luminosity"] += (
                4.0
                * np.pi
                * (model_box.parameters["disk_radius"] * constants.R_JUP) ** 2
                * constants.SIGMA_SB
                * model_box.parameters["disk_teff"] ** 4.0
                / constants.L_SUN
            )  # (Lsun)

        return model_box

    @typechecked
    def get_flux(
        self, model_param: Dict[str, float], synphot=None, return_box: bool = False
    ) -> Union[Tuple[Optional[float], Optional[float]], PhotometryBox]:
        """
        Function for calculating the average flux density for the
        ``filter_name``.

        Parameters
        ----------
        model_param : dict
            Model parameters and values.
        synphot : SyntheticPhotometry, None
            Synthetic photometry object. The object is created if set
            to ``None``.
        return_box : bool
            Return a :class:`~species.core.box.PhotometryBox`
            if set to ``True`` or return the two values that are
            specified below if set to ``False``. By default, the
            argument is set to ``False``. The advantage of
            returning the output in a
            :class:`~species.core.box.PhotometryBox` is that it can
            directly be provided as input to
            :func:`~species.plot.plot_spectrum.plot_spectrum`.

        Returns
        -------
        float
            Average flux (W m-2 um-1).
        float, None
            Uncertainty (W m-2 um-1), which is set to ``None``.
        """

        param_list = self.get_parameters()

        for key in param_list:
            if key not in model_param.keys():
                raise ValueError(
                    f"The '{key}' parameter is required by '{self.model}'. "
                    f"The mandatory parameters are {param_list}."
                )

        if self.spectrum_interp is None:
            self.interpolate_model()

        spectrum = self.get_model(model_param)

        if synphot is None:
            synphot = SyntheticPhotometry(self.filter_name)

        model_flux = synphot.spectrum_to_flux(spectrum.wavelength, spectrum.flux)

        if return_box:
            model_mag = self.get_magnitude(model_param)

            phot_box = create_box(
                boxtype="photometry",
                name=self.model,
                wavelength=[self.mean_wavelength],
                flux=[model_flux],
                app_mag=[(model_mag[0], None)],
                abs_mag=[(model_mag[1], None)],
                filter_name=[self.filter_name],
            )

            return phot_box

        return model_flux

    @typechecked
    def get_magnitude(
        self,
        model_param: Dict[str, float],
        return_box: bool = False,
    ) -> Union[Tuple[Optional[float], Optional[float]], PhotometryBox]:
        """
        Function for calculating the apparent and absolute magnitudes
        for the ``filter_name``.

        Parameters
        ----------
        model_param : dict
            Dictionary with the model parameters. A ``radius`` (Rjup),
            and ``parallax`` (mas) or ``distance`` (pc) are required
            for the apparent magnitude (i.e. to scale the flux from
            the planet to the observer). Only a ``radius`` is
            required for the absolute magnitude.
        return_box : bool
            Return a :class:`~species.core.box.PhotometryBox`
            if set to ``True`` or return the two values that are
            specified below if set to ``False``. By default, the
            argument is set to ``False``. The advantage of
            returning the output in a
            :class:`~species.core.box.PhotometryBox` is that it can
            directly be provided as input to
            :func:`~species.plot.plot_spectrum.plot_spectrum`.

        Returns
        -------
        float
            Apparent magnitude. A ``None`` is returned if the
            dictionary of ``model_param`` does not contain a
            ``radius``, and ``parallax`` or ``distance``.
        float, None
            Absolute magnitude. A ``None`` is returned if the
            dictionary of ``model_param`` does not contain a
            ``radius``.
        """

        param_list = self.get_parameters()

        for key in param_list:
            if key not in model_param.keys():
                raise ValueError(
                    f"The '{key}' parameter is required by '{self.model}'. "
                    f"The mandatory parameters are {param_list}."
                )

        if self.spectrum_interp is None:
            self.interpolate_model()

        try:
            spectrum = self.get_model(model_param)

        except ValueError:
            warnings.warn(
                f"The set of model parameters {model_param} is outside "
                f"the grid range {self.get_bounds()} so returning a NaN."
            )

            return np.nan, np.nan

        if spectrum.wavelength.size == 0:
            app_mag = (np.nan, None)
            abs_mag = (np.nan, None)

        else:
            synphot = SyntheticPhotometry(self.filter_name)

            if "radius" in model_param and "parallax" in model_param:
                app_mag, abs_mag = synphot.spectrum_to_magnitude(
                    spectrum.wavelength,
                    spectrum.flux,
                    parallax=(model_param["parallax"], None),
                )

            elif "radius" in model_param and "distance" in model_param:
                app_mag, abs_mag = synphot.spectrum_to_magnitude(
                    spectrum.wavelength,
                    spectrum.flux,
                    distance=(model_param["distance"], None),
                )

            else:
                app_mag = (None, None)
                abs_mag = (None, None)

                if "radius" in model_param:
                    distance = 10.0  # (pc)

                    spectrum.flux *= (model_param["radius"] * constants.R_JUP) ** 2
                    spectrum.flux /= (distance * constants.PARSEC) ** 2

                    _, abs_mag = synphot.spectrum_to_magnitude(
                        spectrum.wavelength, spectrum.flux, distance=(distance, None)
                    )

        if return_box:
            model_flux = self.get_flux(model_param)

            phot_box = create_box(
                boxtype="photometry",
                name=self.model,
                wavelength=[self.mean_wavelength],
                flux=[model_flux],
                app_mag=[(app_mag[0], None)],
                abs_mag=[(abs_mag[0], None)],
                filter_name=[self.filter_name],
            )

            return phot_box

        return app_mag[0], abs_mag[0]

    @typechecked
    def get_bounds(self) -> Dict[str, Tuple[float, float]]:
        """
        Function for extracting the grid boundaries.

        Returns
        -------
        dict
            Boundaries of parameter grid.
        """

        param_list = self.get_parameters()

        with self.open_database() as hdf5_file:
            bounds = {}

            for param_item in param_list:
                data = hdf5_file[f"models/{self.model}/{param_item}"]
                bounds[param_item] = (data[0], data[-1])

        return bounds

    @typechecked
    def get_wavelengths(self) -> np.ndarray:
        """
        Function for extracting the wavelength points.

        Returns
        -------
        np.ndarray
            Wavelength points (um).
        """

        with self.open_database() as hdf5_file:
            wavelength = np.array(hdf5_file[f"models/{self.model}/wavelength"])

        return wavelength

    @typechecked
    def get_points(self) -> Dict[str, np.ndarray]:
        """
        Function for extracting the grid points.

        Returns
        -------
        dict
            Parameter points of the model grid.
        """

        param_list = self.get_parameters()

        points = {}

        with self.open_database() as hdf5_file:
            points = {}

            for param_item in param_list:
                data = hdf5_file[f"models/{self.model}/{param_item}"]
                points[param_item] = np.array(data)

        return points

    @typechecked
    def get_parameters(self) -> List[str]:
        """
        Function for extracting the parameter names.

        Returns
        -------
        list(str)
            Model parameters.
        """

        with self.open_database() as hdf5_file:
            dset = hdf5_file[f"models/{self.model}"]

            if "n_param" in dset.attrs:
                n_param = dset.attrs["n_param"]

            elif "nparam" in dset.attrs:
                n_param = dset.attrs["nparam"]

            param = []
            for i in range(n_param):
                param.append(dset.attrs[f"parameter{i}"])

        return param

    @typechecked
    def get_sampling(self) -> float:
        """
        Function for returning the wavelength sampling,
        :math:`\\lambda/\\Delta\\lambda`, of the
        model spectra as stored in the database.

        Returns
        -------
        float
            Wavelength sampling, :math:`\\lambda/\\Delta\\lambda`.
        """

        wavel_points = self.get_wavelengths()

        wavel_mean = (wavel_points[1:] + wavel_points[:-1]) / 2.0
        wavel_sampling = wavel_mean / np.diff(wavel_points)

        return np.mean(wavel_sampling)

    @typechecked
    def binary_spectrum(
        self,
        model_param: Dict[str, float],
        spec_res: Optional[float] = None,
        wavel_resample: Optional[np.ndarray] = None,
        **kwargs,
    ) -> ModelBox:
        """
        Function for extracting a model spectrum of a binary system.
        A weighted combination of two spectra will be returned. The
        ``model_param`` dictionary should contain the parameters
        for both components (e.g. ``teff_0`` and ``teff_1``, instead
        of ``teff``). Apart from that, the same parameters are used
        as with :meth:`~species.read.read_model.ReadModel.get_model`.

        Parameters
        ----------
        model_param : dict
            Dictionary with the model parameters and values. The values
            should be within the boundaries of the grid. The grid
            boundaries of the spectra in the database can be obtained
            with
            :func:`~species.read.read_model.ReadModel.get_bounds()`.
        spec_res : float, None
            Spectral resolution that is used for smoothing the
            spectrum with a Gaussian kernel. No smoothing is applied
            if the argument is set to ``None``.
        wavel_resample : np.ndarray, None
            Wavelength points (um) to which the spectrum is resampled.
            In that case, ``spec_res`` can still be used for smoothing
            the spectrum with a Gaussian kernel. The original
            wavelength points are used if the argument is set to
            ``None``.

        Returns
        -------
        species.core.box.ModelBox
            Box with the model spectrum.
        """

        if "smooth" in kwargs:
            warnings.warn(
                "The 'smooth' parameter has been "
                "deprecated. Please set only the "
                "'spec_res' argument, which can be set "
                "to None for not applying a smoothing.",
                DeprecationWarning,
            )

            if not kwargs["smooth"] and spec_res is not None:
                spec_res = None

        # Get grid boundaries

        param_0 = binary_to_single(model_param, 0)

        model_box_0 = self.get_model(
            param_0,
            spec_res=spec_res,
            wavel_resample=wavel_resample,
        )

        param_1 = binary_to_single(model_param, 1)

        model_box_1 = self.get_model(
            param_1,
            spec_res=spec_res,
            wavel_resample=wavel_resample,
        )

        flux_comb = (
            model_param["spec_weight"] * model_box_0.flux
            + (1.0 - model_param["spec_weight"]) * model_box_1.flux
        )

        model_box = create_box(
            boxtype="model",
            model=self.model,
            wavelength=model_box_0.wavelength,
            flux=flux_comb,
            parameters=model_param,
            quantity="flux",
            spec_res=spec_res,
        )

        return model_box

    @typechecked
    def integrate_spectrum(self, model_param: Dict[str, float]) -> float:
        """
        Function for calculating the bolometric flux by integrating
        a model spectrum at the requested parameters. In principle,
        the calculated luminosity should be approximately the same
        as the value that can be calculated directly from the
        :math:`T_\\mathrm{eff}` and radius parameters, unless the
        atmospheric model had not properly converged.

        Parameters
        ----------
        model_param : dict
            Dictionary with the model parameters and values. The values
            should be within the boundaries of the grid. The grid
            boundaries of the spectra in the database can be obtained
            with
            :func:`~species.read.read_model.ReadModel.get_bounds()`.

        Returns
        -------
        float
            Bolometric luminosity (:math:`\\log{(L/L_\\odot)}`).
        """

        if "radius" not in model_param:
            raise ValueError(
                "Please include the 'radius' parameter "
                "in the 'model_param' dictionary, "
                "which is require for calculating the "
                "bolometric luminosity."
            )

        wavel_points = self.get_wavelengths()

        if self.wavel_range is None:
            self.wavel_range = (wavel_points[0], wavel_points[-1])

        if (
            self.wavel_range[0] != wavel_points[0]
            or self.wavel_range[1] != wavel_points[-1]
        ):
            warnings.warn(
                "The 'wavel_range' is not set to the "
                "maximum available range. To maximize the "
                "accuracy when calculating the bolometric "
                "luminosity, it is recommended to set "
                "'wavel_range=None'."
            )

        if "parallax" in model_param:
            del model_param["parallax"]

        if "distance" in model_param:
            del model_param["distance"]

        model_box = self.get_model(model_param)

        bol_lum = (
            4.0
            * np.pi
            * (model_param["radius"] * constants.R_JUP) ** 2
            * simps(model_box.flux, model_box.wavelength)
        )

        return np.log10(bol_lum / constants.L_SUN)

    @typechecked
    def create_color_magnitude(
        self,
        model_param: Dict[str, float],
        filters_color: Tuple[str, str],
        filter_mag: str,
    ) -> ColorMagBox:
        """
        Function for creating a :class:`~species.core.box.
        ColorMagBox` for a given set of filter names and model
        parameters. The effective temperature, :math:`T_\\mathrm{eff}`,
        is varied such that the returned :class:`~species.core.box.
        ColorMagBox` contains the colors as function of
        :math:`T_\\mathrm{eff}` and can be provide as input to the
        :func:`~species.plot.plot_color.plot_color_magnitude` function.

        Parameters
        ----------
        model_param : dict
            Dictionary with the model parameters and values. The values
            should be within the boundaries of the grid. The boundaries
            of the model grid can be inspected by using the
            :func:`~species.read.read_model.ReadModel.get_bounds()`
            method. The effective temperature, :math:`T_\\mathrm{eff}`,
            does not need to be included in the dictionary since it
            is varied. The values of :math:`T_\\mathrm{eff}` are set to
            the grid points. The grid points can be inspected with the
            :func:`~species.read.read_model.ReadModel.get_points()`
            method.
        filters_color : tuple(str, str)
            Filter names that are used for the color. Any of
            the filter names from the `SVO Filter Profile Service
            <http://svo2.cab.inta-csic.es/svo/theory/fps/>`_ are
            compatible.
        filter_mag : str
            Filter name that is used for the magnitude. Any of
            the filter names from the `SVO Filter Profile Service
            <http://svo2.cab.inta-csic.es/svo/theory/fps/>`_ are
            compatible.

        Returns
        -------
        species.core.box.ColorMagBox
            Box with the colors and magnitudes.
        """

        if "distance" not in model_param:
            model_param["distance"] = 10.0

        if "radius" not in model_param:
            model_param["radius"] = 1.0

        if "parallax" in model_param:
            del model_param["parallax"]

        read_model_1 = ReadModel(self.model, filter_name=filters_color[0])
        read_model_2 = ReadModel(self.model, filter_name=filters_color[0])
        read_model_3 = ReadModel(self.model, filter_name=filter_mag)

        model_points = self.get_points()

        param_list = []
        color_list = []
        mag_list = []

        for param_item in model_points["teff"]:
            model_param["teff"] = param_item

            mag_1 = read_model_1.get_magnitude(model_param)
            mag_2 = read_model_2.get_magnitude(model_param)
            mag_3 = read_model_3.get_magnitude(model_param)

            param_list.append(param_item)
            color_list.append(mag_1[0] - mag_2[0])
            mag_list.append(mag_3[0])

        return create_box(
            "colormag",
            library=self.model,
            object_type="spectra",
            filters_color=filters_color,
            filter_mag=filter_mag,
            color=color_list,
            magnitude=mag_list,
            sptype=param_list,
        )

    @typechecked
    def create_color_color(
        self,
        model_param: Dict[str, float],
        filters_colors: Tuple[Tuple[str, str], Tuple[str, str]],
    ) -> ColorColorBox:
        """
        Function for creating a :class:`~species.core.box.
        ColorColorBox` for a given set of filter names and model
        parameters. The effective temperature, :math:`T_\\mathrm{eff}`,
        is varied such that the returned :class:`~species.core.box.
        ColorColorBox` contains the colors as function of
        :math:`T_\\mathrm{eff}` and can be provide as input to the
        :func:`~species.plot.plot_color.plot_color_color` function.

        Parameters
        ----------
        model_param : dict
            Dictionary with the model parameters and values. The values
            should be within the boundaries of the grid. The boundaries
            of the model grid can be inspected by using the
            :func:`~species.read.read_model.ReadModel.get_bounds()`
            method. The effective temperature, :math:`T_\\mathrm{eff}`,
            does not need to be included in the dictionary since it
            is varied. The values of :math:`T_\\mathrm{eff}` are set to
            the grid points. The grid points can be inspected with the
            :func:`~species.read.read_model.ReadModel.get_points()`
            method.
        filters_colors : tuple(tuple(str, str), tuple(str, str))
            Filter names that are used for the two colors. Any of
            the filter names from the `SVO Filter Profile Service
            <http://svo2.cab.inta-csic.es/svo/theory/fps/>`_ are
            compatible.

        Returns
        -------
        species.core.box.ColorColorBox
            Box with the colors.
        """

        if "distance" not in model_param:
            model_param["distance"] = 10.0

        if "radius" not in model_param:
            model_param["radius"] = 1.0

        if "parallax" in model_param:
            del model_param["parallax"]

        read_model_1 = ReadModel(self.model, filter_name=filters_colors[0][0])
        read_model_2 = ReadModel(self.model, filter_name=filters_colors[0][1])
        read_model_3 = ReadModel(self.model, filter_name=filters_colors[1][0])
        read_model_4 = ReadModel(self.model, filter_name=filters_colors[1][1])

        model_points = self.get_points()

        param_list = []
        color_1_list = []
        color_2_list = []

        for param_item in model_points["teff"]:
            model_param["teff"] = param_item

            mag_1 = read_model_1.get_magnitude(model_param)
            mag_2 = read_model_2.get_magnitude(model_param)
            mag_3 = read_model_3.get_magnitude(model_param)
            mag_4 = read_model_4.get_magnitude(model_param)

            param_list.append(param_item)
            color_1_list.append(mag_1[0] - mag_2[0])
            color_2_list.append(mag_3[0] - mag_4[0])

        return create_box(
            "colorcolor",
            library=self.model,
            object_type="spectra",
            filters=filters_colors,
            color1=color_1_list,
            color2=color_2_list,
            sptype=param_list,
        )



"""
Module for adding O5 through L3 SDSS stellar spectra from
Kesseli et al. (2017) to the database.
"""

import os
import shutil
import urllib.request

import h5py
import numpy as np

from astropy.io import fits
from typeguard import typechecked

from species.util.data_util import extract_tarfile


@typechecked
def add_kesseli2017(input_path: str, database: h5py._hl.files.File) -> None:
    """
    Function for adding the SDSS stellar spectra from
    Kesseli et al. (2017) to the database.

    Parameters
    ----------
    input_path : str
        Path of the data folder.
    database : h5py._hl.files.File
        The HDF5 database.

    Returns
    -------
    NoneType
        None
    """

    data_url = "https://cdsarc.unistra.fr/viz-bin/nph-Cat/tar.gz?J/ApJS/230/16"
    data_file = os.path.join(input_path, "J_ApJS_230_16.tar.gz")
    data_folder = os.path.join(input_path, "kesseli+2017/")

    if not os.path.isfile(data_file):
        print(
            "Downloading SDSS spectra from Kesseli et al. 2017 (145 MB)...",
            end="",
            flush=True,
        )
        urllib.request.urlretrieve(data_url, data_file)
        print(" [DONE]")

    if os.path.exists(data_folder):
        shutil.rmtree(data_folder)

    print(
        "Unpacking SDSS spectra from Kesseli et al. 2017 (145 MB)...",
        end="",
        flush=True,
    )
    extract_tarfile(data_file, data_folder)
    print(" [DONE]")

    database.create_group("spectra/kesseli+2017")

    fits_folder = os.path.join(data_folder, "fits")

    print_message = ""

    for _, _, files in os.walk(fits_folder):
        for _, filename in enumerate(files):
            with fits.open(os.path.join(fits_folder, filename)) as hdu_list:
                data = hdu_list[1].data

                wavelength = 1e-4 * 10.0 ** data["LogLam"]  # (um)
                flux = data["Flux"]  # Normalized units
                error = data["PropErr"]  # Normalized units

                name = filename[:-5].replace("_", " ")

                file_split = filename.split("_")
                file_split = file_split[0].split(".")

                sptype = file_split[0]

                spdata = np.column_stack([wavelength, flux, error])

                empty_message = len(print_message) * " "
                print(f"\r{empty_message}", end="")

                print_message = f"Adding spectra... {name}"
                print(f"\r{print_message}", end="")

                dset = database.create_dataset(
                    f"spectra/kesseli+2017/{name}", data=spdata
                )

                dset.attrs["name"] = str(name).encode()
                dset.attrs["sptype"] = str(sptype).encode()

    empty_message = len(print_message) * " "
    print(f"\r{empty_message}", end="")

    print_message = "Adding spectra... [DONE]"
    print(f"\r{print_message}")
